import unittest
import numpy as np
from pydarm.plot import plot, residuals
from pydarm.darm import DARMModel
from pydarm.measurement import Measurement


class TestDarmPlot(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 100)

    def tearDown(self):
        del self.frequencies

    def test_plot(self):
        darm = DARMModel('''
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_uncompensated_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1638.001638001638, 1638.001638001638
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules = 4: 4
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_test = ON, ON

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
''')
        darm.plot(label=['First model'],
             show=False)


class TestPlotPlot(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 100)

    def tearDown(self):
        del self.frequencies

    def test_plot(self):
        darm = DARMModel('''
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_uncompensated_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1638.001638001638, 1638.001638001638
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules = 4: 4
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_test = ON, ON

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
''')
        test_olg = darm.compute_darm_olg(self.frequencies)
        plot(self.frequencies, test_olg,
             title='Test Plot',
             label=['DARM OLG'],
             show=False)


class TestResiduals(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(0, np.log10(5000.), 100)

    def tearDown(self):
        del self.frequencies

    def test_plot(self):

        meas_object = Measurement(
            './test/2020-01-03_H1_DARM_OLGTF_LF_SS_5to1100Hz_15min.xml')
        meas_freq, meas_tf, meas_coh, meas_unc = meas_object.get_raw_tf(
            'H1:LSC-DARM1_IN2', 'H1:LSC-DARM1_IN1')
        darm = DARMModel('''
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_uncompensated_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1638.001638001638, 1638.001638001638
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules = 4: 4
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_test = ON, ON

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
''')
        test_olg = darm.compute_darm_olg(self.frequencies)
        residuals(self.frequencies, -test_olg,
                  meas_freq, meas_tf, meas_unc,
                  plot_input=True,
                  show=False)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) Jameson Rollins (2021)
#               Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import os
import sys
import signal
import logging
import argparse
from importlib import import_module

from . import __version__
from .cmd import CMDS, CMDError, logger


logger.setLevel(os.getenv('LOG_LEVEL', 'INFO').upper())
formatter = logging.Formatter('%(message)s')
handler = logging.StreamHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)

########################################

parser = argparse.ArgumentParser(
    prog='pydarm',
    description="""aLIGO calibration interface

""",
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
parser.add_argument(
    '--version', '-v', action='version', version=__version__)
subparsers = parser.add_subparsers(
    metavar='COMMAND',
)


def add_subcommand(name):
    """helper function for adding subcommand to the parser from cmd modules

    The name provided should be the name of a module in the `cmd`
    sub-package.  Each module should include an `add_args` function
    for adding argparse arguments to the argparse parser, and a `main`
    function with the main subcommand logic.  The docstring of the
    main function will be the doc for the sub-command.

    """
    mod = import_module(f'.{name}', 'pydarm.cmd')
    func = mod.main
    sp = subparsers.add_parser(
        name,
        help=func.__doc__.splitlines()[0],
        description=func.__doc__.strip(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    sp.set_defaults(func=func)
    mod.add_args(sp)
    return sp


def main():
    for name in CMDS:
        add_subcommand(name)
    args = parser.parse_args()
    if 'func' not in args:
        parser.print_help()
        parser.exit()
    func = args.func
    del args.func
    logger.debug(args)
    func(args)


if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    try:
        main()
    except CMDError as e:
        sys.exit(f"ERROR: {e}")



# Copyright (C) Jameson Rollins (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
from matplotlib.backends.backend_pdf import PdfPages
from scipy.interpolate import interp1d

from . import darm


plt.rcParams.update({'text.usetex': False,
                     'lines.linewidth': 3,
                     'font.family': 'sans-serif',
                     'font.serif': 'Helvetica',
                     'font.size': 10,
                     'xtick.labelsize': 'x-large',
                     'xtick.direction': 'in',
                     'ytick.labelsize': 'x-large',
                     'ytick.direction': 'in',
                     'axes.labelsize': 'medium',
                     'axes.titlesize': 'x-large',
                     'axes.grid': True,
                     'grid.alpha': 0.5,
                     'lines.markersize': 12,
                     'legend.borderpad': 0.2,
                     'legend.fancybox': True,
                     'legend.fontsize': 'medium',
                     'legend.framealpha': 0.7,
                     'legend.handletextpad': 0.1,
                     'legend.labelspacing': 0.2,
                     'legend.loc': 'best',
                     'figure.figsize': (12, 8),
                     'savefig.dpi': 100,
                     'savefig.orientation': 'landscape',
                     'pdf.compression': 9})


class BodePlot:
    """bode plotting object

    """
    def __init__(self, fig=None, title=None, figsize=(12, 8), only='', spspec=[211, 212]):
        """
        Parameters
        ----------
        only (string): if set to 'mag' or 'phase', this object will only
        plot magnitude or phase, and not the usual 2x1 body plot.

        Notes
        -----

        """
        self.fig = fig or plt.figure(figsize=figsize)
        self.only = only

        self.phase_yticks = np.arange(-180, 180+30, 30)
        self.phase_ylim = (-185, 185)

        if self.only == '':
            self.ax_mag = self.fig.add_subplot(spspec[0])
            self.ax_mag.grid(True)
            self.ax_mag.grid(which='minor', axis='both', linestyle='--')
            # self.ax_mag.axhline(1, color='k', linestyle='dashed')
            self.ax_mag.set_ylabel('Magnitude')

            self.ax_phase = self.fig.add_subplot(spspec[1])
            self.ax_phase.grid(True)
            self.ax_phase.grid(which='minor', axis='both', linestyle='--')
            self.ax_phase.set_yticks(self.phase_yticks)
            self.ax_phase.set_ylabel('Phase [deg]')
            self.ax_phase.set_xlabel('Frequency [Hz]')
            self.ax_phase.set_ylim(self.phase_ylim)

        if self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag = self.fig.add_subplot(111)
            self.ax_mag.grid(True)
            self.ax_mag.grid(which='minor', axis='both', linestyle='--')
            # self.ax_mag.axhline(1, color='k', linestyle='dashed')
            self.ax_mag.set_ylabel('Magnitude')
            self.ax_mag.set_xlabel('Frequency [Hz]')

        if self.only == 'phase':
            self.ax_phase = self.fig.add_subplot(111)
            self.ax_phase.grid(True)
            self.ax_phase.grid(which='minor', axis='both', linestyle='--')
            self.ax_phase.set_yticks(self.phase_yticks)
            self.ax_phase.set_ylabel('Phase [deg]')
            self.ax_phase.set_xlabel('Frequency [Hz]')
            self.ax_phase.set_ylim(self.phase_ylim)

        if title:
            self.fig.suptitle(title, y=0.9)

    def plot(self, freq, tf, **kwargs):
        """add frequency response to plot

        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag.loglog(freq, np.abs(tf), **kwargs)
        if self.only == '' or self.only == 'phase':
            self.ax_phase.semilogx(freq, np.angle(tf, deg=True), **kwargs)

    def error(self, freq, tf, tf_error, **kwargs):
        """add frequency response to plot with error
           Note: Error should be relative.

        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag.errorbar(freq, np.abs(tf), np.abs(tf)*tf_error, **kwargs)
            self.ax_mag.set_xscale('log')
            self.ax_mag.set_yscale('log')
        if self.only == '' or self.only == 'phase':
            self.ax_phase.errorbar(freq, np.angle(tf, deg=True), tf_error*180/np.pi, **kwargs)
            self.ax_phase.set_xscale('log')
            self.ax_phase.set_yscale('linear')
            self.ax_phase.set_yticks(self.phase_yticks)

    def legend(self, labels=None, **kwargs):
        """add legend
        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            if labels:
                self.ax_mag.legend(labels, **kwargs)
            else:  # labels can also be passed as kwargs to plot, in which case this is fine
                self.ax_mag.legend(**kwargs)
        if self.only == 'phase':
            if labels:
                self.ax_mag.legend(labels, **kwargs)
            else:
                self.ax_phase.legend(**kwargs)

    def save(self, path):
        """save bode plot to file

        """

        self.fig.savefig(path)

    def show(self):
        """show interactive plot

        """

        plt.show()

    def xlim(self, freq_min, freq_max):
        """set min max for frequency axis

        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag.set_xlim(freq_min, freq_max)
            self.autoscale_mag_y()
        if self.only == '' or self.only == 'phase':
            self.ax_phase.set_xlim(freq_min, freq_max)

    def ylim(self, mag_min, mag_max, phase_min=-185, phase_max=185):
        """set min max for the magnitude (and phase) axis

        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag.set_ylim(mag_min, mag_max)
        if self.only == '' or self.only == 'phase':
            self.ax_phase.set_ylim(phase_min, phase_max)

    def vlines(self, vfreq, color='red', lw=1.0, linestyle='--', label="_nolegend_"):
        """add vertical line at specific frequency in both mag and phase plot

        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag.axvline(vfreq, color=color, lw=lw, linestyle=linestyle, label=label)
        if self.only == '' or self.only == 'phase':
            self.ax_phase.axvline(vfreq, color=color, lw=lw, linestyle=linestyle, label=label)

    def text(self, x, y, plot_text):
        """add text in plot

        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag.text(x, y, plot_text)
        if self.only == 'phase':
            self.ax_phase.text(x, y, plot_text)

    def autoscale_mag_y(self, margin=0.1):
        """This function rescales the y-axis based on the data that is visible given
        the current xlim of the axis.

        """

        def get_bottom_top(line):
            xd = np.array(line.get_xdata())
            yd = np.array(line.get_ydata())
            lo, hi = self.ax_mag.get_xlim()
            y_displayed = yd[((xd > lo) & (xd < hi))]
            bot = np.min(np.abs(y_displayed[np.nonzero(y_displayed)]))*(1-margin)
            top = np.max(y_displayed)*(1+margin)
            return bot, top

        lines = self.ax_mag.get_lines()
        bot, top = np.inf, -np.inf

        for line in lines:
            new_bot, new_top = get_bottom_top(line)
            if new_bot < bot:
                bot = new_bot
            if new_top > top:
                top = new_top
        self.ax_mag.set_ylim(bot, top)

    def greed(self, more_minor=True):
        """add a finer grid than the default

        """

        if self.only == '' or self.only == 'mag' or self.only == 'magnitude':
            self.ax_mag.yaxis.set_major_locator(plticker.LogLocator(base=10, numticks=200))
            if more_minor:
                locmin = plticker.LogLocator(base=10,
                                             subs=(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9),
                                             numticks=6000)
                self.ax_mag.yaxis.set_minor_locator(locmin)
            self.ax_mag.grid(which='major', axis='both', linestyle='-')

        if self.only == '' or self.only == 'phase':
            self.ax_phase.yaxis.set_major_locator(plticker.MultipleLocator(base=45))
            self.ax_phase.yaxis.set_minor_locator(plticker.MultipleLocator(base=5))
            self.ax_phase.grid(which='major', axis='both', linestyle='-')


def plot(*args, freq_min=0.1, freq_max=5000, freq_points=1001, filename=None,
         mag_min=None, mag_max=None, phase_min=-185, phase_max=185, label=None,
         title=None, style=None, show=None, greed=False, figsize=(12, 8), **kwargs):
    """
    This method provides a simple way to plot quickly anything you want.

    Parameters
    ----------
    args : tuple, lists
        Here you can input any kind of transfer function related data.
        It accepts multiple sets in multiple formats.
        FORMATS:
        - 2-tuple (freq,transfer function)
        - 3-tuple (freq,transfer function, rel unc)
        - 4-tuple (freq,transfer function, coherence, rel unc)
        - freq, transfer function (two lists or 1d arrays separated by comma)
        - method that returns a transfer function e.g. D.compute_darm_olg
        (assuming D is a darm model object)
    freq_min : `float`, optional
        start frequency
    freq_max : `float`, optional
        end frequency
    freq_points : `int`, optional
        number of points to calculate
    filename : `str`, optional
        if given, ALL generated graphs will be saved in one pdf.
    mag_min : `float`, optional
        minimum magnitude scale
    mag_max : `float`, optional
        maximum magnitude scale
    phase_min : `float`, optional
        minimum phase scale
    phase_max : `float`, optional
        maximum phase scale
    label : `str` list, optional
        legend string for the data sets given e.g. ['first data', 'second']
    title : `str`, optional
        title of the graph
    style : `str`, optional
        one of the styles matplotlib has or a user filename with style
    show : `bool`, optional
        if true the plot will show
    greed : `bool`, optional
        if true a finer grid is drawn for all figs
    figsize : `float` tuple, optional
        figure size
    **kwargs
        matplotlib arguments passed to all plots

    Returns
    -------
    bp : `pydarm.plot.BodePlot` object

    """

    if filename:
        plt.close('all')
    if style:
        plt.style.use(style)

    # BodyPlot object.
    bp = BodePlot(title=title, figsize=figsize)

    # Temporary index used to identify frequency inputs and transfer function inputs.
    temp_pair_index = 0

    for argument in args:
        # Check if a simple list. If so, then this assumes it is frequency
        # The next one will be taken to be a TF
        if isinstance(argument, (np.ndarray, list)):
            temp_pair_index = temp_pair_index+1
            if (temp_pair_index % 2) == 1:
                # This array should be the frerquency array
                freq_in = argument
            if (temp_pair_index % 2) == 0:
                # This array should then be a TF array
                tf_in = argument
                bp.plot(freq_in, tf_in, **kwargs)
        # Check if this is a tuple. If so, a tuple of size 2 is assumed to
        # be (freq,tf) pair, and if a tuple of size 4 is assumed to be a TF
        # that comes from the measurement class with (freq,tf,coh,unc).
        elif isinstance(argument, (tuple)):
            if len(argument) == 2:
                # This tuple should be (freq,tf)
                freq_in = argument[0]
                tf_in = argument[1]
                bp.plot(freq_in, tf_in, **kwargs)
            elif len(argument) == 3:
                # This tuple should be (freq,tf,unc)
                freq_in = argument[0]
                tf_in = argument[1]
                # Uncertainty is relative, and needs to be for error plot.
                tf_unc = argument[2]
                bp.error(freq_in, tf_in, tf_unc, fmt='ro', linestyle='None', **kwargs)
            elif len(argument) == 4:
                # This tuple should be (freq,tf,coh,unc)
                freq_in = argument[0]
                tf_in = argument[1]
                # tf_coh = argument[2]
                # Uncertainty is relative, and needs to be for error plot.
                tf_unc = argument[3]
                bp.error(freq_in, tf_in, tf_unc, fmt='ro', linestyle='None', **kwargs)
            else:
                raise ValueError('Bad tuple given for plotting')
        else:
            # This argument is a method. We use the frequencies given
            # or the default ones.
            freq = np.logspace(np.log10(float(freq_min)),
                               np.log10(float(freq_max)), int(freq_points))
            tf_in = argument(freq)
            bp.plot(freq, tf_in, **kwargs)

    if label:
        bp.legend(label)
    bp.xlim(freq_min, freq_max)
    if greed:
        bp.greed()
    if mag_min:
        bp.ylim(mag_min, mag_max, phase_min=phase_min, phase_max=phase_max)
    if filename:
        bp.save(filename)
    if show:
        plt.show()
    # Return the handle of the bode plot in case you want to do more things
    # to your plot. Does not return handles of critiques.
    return bp


def residuals(freq_model, tf_model, freq_data, tf_data, unc_data,
              mag_min=0.8, mag_max=1.2, phase_min=-5, phase_max=5,
              mag_major_tick=0.05, mag_minor_tick=0.01,
              phase_major_tick=2, phase_minor_tick=0.5,
              filename=None, show=None, plot_input=False):
    """
    This method provides a simple way to plot quickly anything you want.

    Parameters
    ----------
    freq_model : `float`, array-like
        frequencies that the model transfer function is computed
    tf_model : `complex`, array-like
        transfer function response values at the `freq_model` points
    freq_data : `float`, array-like
        frequencies that the measured transfer function data values are
        computed at
    tf_data : `complex`, array-like
        measured transfer function response values at the `freq_daata` points
    unc_data : `float`, array-like
        relative uncertainty values, determined from the measurement coherence
    mag_min : `float`, optional
        minimum magnitude scale
    mag_max : `float`, optional
        maximum magnitude scale
    phase_min : `float`, optional
        minimum phase scale
    phase_max : `float`, optional
        maximum phase scale
    mag_major_tick : `float`, optional
        major tick placement for the magnitude plot
    mag_minor_tick : `float`, optional
        minor tick placement for the magnitude plot
    phase_major_tick : `float`, optional
        major tick placement for the phase plot
    phase_minor_tick : `float`, optional
        minor tick placement for the phase plot
    filename : `str`, optional
        save the plot to this file
    show : `bool`, optional
        if true the plot will show
    plot_input : `bool`, optional
        FIXME: what does this do?

    Notes
    -----
    Example usage:
    >>> measurement_tuple = freq_data, tf_data, unc_data
    >>> plot(freq_model, tf_model, measurement_tuple,
    ... freq_min=freq_data[0], freq_max=freq_data[-1],
    ... title='Measurement vs Model')

    """

    # First interpolate the model tf at the frequencies of the data points
    tf_model_interp = np.interp(freq_data, freq_model, tf_model)
    tf_ratio = tf_data/tf_model_interp
    tf_ratio_tuple = freq_data, tf_ratio, unc_data

    if plot_input:
        if filename is not None:
            filename_input = filename[0:-4]+'_input'+filename[len(filename)-4:len(filename)]
        else:
            filename_input = None
        plot(freq_model, tf_model, (freq_data, tf_data, unc_data), title='Measurement vs Model',
             freq_min=freq_data[0], freq_max=freq_data[-1], filename=filename_input)

    bode_handle = plot(tf_ratio_tuple, freq_min=freq_data[0], freq_max=freq_data[-1],
                       mag_min=mag_min, mag_max=mag_max, phase_min=phase_min,
                       phase_max=phase_max, show=show, title='Measurement vs Model Residuals')
    bode_handle.ax_mag.set_yscale('linear')

    bode_handle.ax_mag.yaxis.set_major_locator(plticker.MultipleLocator(base=mag_major_tick))
    bode_handle.ax_mag.yaxis.set_minor_locator(plticker.MultipleLocator(base=mag_minor_tick))
    bode_handle.ax_mag.grid(which='major', axis='both', linestyle='-')
    bode_handle.ax_phase.yaxis.set_major_locator(plticker.MultipleLocator(base=phase_major_tick))
    bode_handle.ax_phase.yaxis.set_minor_locator(plticker.MultipleLocator(base=phase_minor_tick))
    bode_handle.ax_phase.grid(which='major', axis='both', linestyle='-')

    if filename:
        bode_handle.save(filename)


def get_ugf(freq, tf, search_start=5., search_end=1000., n_points=10000):
    """
    Very non-fancy function for easily finding UGF in the simplest of cases.

    Parameters
    ----------
    freq: (float) array of frequencies corresponding to the given tf
    tf: (complex) array of the transfer function you want the ugf of
    search_start: (float) frequency to start UGF search
    search_end: (float) frequency to end UGF search
    n_points: (int) increase for increased accuracy of UGF frequency

    Returns
    -------
    2-tuple float: UGF frequency, phase margin

    """

    mag = np.abs(tf)
    phase = np.angle(tf)*180/np.pi
    f_mag = interp1d(freq, mag, kind='cubic', fill_value="extrapolate")
    f_phase = interp1d(freq, phase, kind='cubic', fill_value="extrapolate")
    freq_new = np.logspace(np.log10(search_start), np.log10(search_end), n_points)
    i = 1
    while f_mag(freq_new[i]) > 1 and i < (n_points-1):
        i = i+1
    if i == n_points-1:
        print('WARNING: UGF NOT FOUND WITHIN THE RANGE GIVEN')
        return -1, 0
    else:
        return np.round(freq_new[i]), np.round(f_phase(freq_new[i]))


def save_multifig(filename, figs=None, dpi=200):
    """method to save all existing (or list of given) figures in one pdf file

    """

    pp = PdfPages(filename)
    if figs is None:
        figs = [plt.figure(n) for n in plt.get_fignums()]
    for fig in figs:
        fig.savefig(pp, format='pdf')
    pp.close()


def critique(*args, filename=None, plot_selection='all', ifo='', show=None,
             label=None, freq_min=0.1, freq_max=5000, freq_points=1001,
             ugf_start=10, ugf_end=1000, greed=True, figsize=(12, 8), **kwargs):
    """
    This method produces critique models for 1 or 2 models.

    Parameters
    ----------
    args : `pydarm.darm.DARMModel` object
        1 or 2 darm models objects, if more than 1, this should be a list
    filename : `str`, optional
        if given, ALL generated graphs will be saved in one pdf
    plot_selection : `str`, optional
        Select plot type, one of: 'all' (default), 'optical', 'actuation',
        'clg', 'olg', or 'digital'
    ifo : `str`, optional
        if given with a model to plot, it will appear in the graph titles
    show : `bool`, optional
        if true the plot(s) will show
    label : `str` list, optional
        FIXME: what should this be?
    freq_min : `float`, optional
        start frequency
    freq_max : `float`, optional
        end frequency
    freq_points : `int`, optional
        Number of frequency points
    ugf_start : `float`, optional
        start frequency used for the search
    ugf_end : `float`, optional
        end frequency used for the search
    greed : `bool`, optional
    figsize : `float`, tuple
    **kwargs : optional
        Matplotlib values passed to plots

    """
    # See bp1 plot
    default_olg_yspan = 1e12
    # See bp2 plot
    default_olg_zoom_fmin = 10
    default_olg_zoom_fmax = 200
    # See bp3b plot
    default_clg_zoom_fmin = 20
    default_clg_zoom_fmax = 400
    # See bp4 plot
    default_dig_yspan = 1e7
    # See bp6 plot
    default_act_strength_yspan = 1e17
    # See bp8, bp11 plot
    default_act_displ_yspan = 1e17

    if filename:
        plt.close('all')

    # Check if all inputs are darm models. If so, it count the number of
    # models given (1 or 2). If not, it complains.
    if all(isinstance(argument, (darm.DARMModel)) for argument in args):
        if len(args) > 2:
            print('You asked critique plots for more than 2 models. Results might be unpleasant.')
    else:
        ValueError('Function plot.critique only takes DARMModel objects as inputs')

    # Temporary index used to identify the first darm model given.
    temp_model_index = 1
    list_of_figures = []

    plt.rc('legend', fontsize=18)

    for argument in args:
        freq = np.logspace(np.log10(float(freq_min)),
                           np.log10(float(freq_max)), int(freq_points))

        G = argument.compute_darm_olg(freq)
        D = argument.digital.compute_response(freq)
        C = argument.sensing.compute_sensing(freq)
        R = argument.compute_response_function(freq)
        # For all actuation calculations I am using param names that
        # (I think) match old pydarm

        # TF from output of actuation digital filter bank
        # to displacement, for each stage
        x_UIM_strength, x_PUM_strength, x_TST_strength = \
            argument.actuation.xarm.digital_out_to_displacement(freq)
        # TF of suspension digital filters
        x_UIM_umodel, x_PUM_umodel, x_TST_umodel = \
            argument.actuation.xarm.sus_digital_filters_response(freq)
        # TF from LOCK_IN to displacement by stage
        x_UIM_authority = \
            argument.actuation.xarm.compute_actuation_single_stage(freq, stage='UIM')
        x_PUM_authority = \
            argument.actuation.xarm.compute_actuation_single_stage(freq, stage='PUM')
        x_TST_authority = \
            argument.actuation.xarm.compute_actuation_single_stage(freq, stage='TST')
        x_ALL_authority = x_UIM_authority+x_PUM_authority+x_TST_authority
        # TF from DARM_IN to discplacement scaled by optical gain
        x_UIM_total = x_UIM_authority*D*argument.sensing.coupled_cavity_optical_gain
        x_PUM_total = x_PUM_authority*D*argument.sensing.coupled_cavity_optical_gain
        x_TST_total = x_TST_authority*D*argument.sensing.coupled_cavity_optical_gain

        # Relative contribution of actuation stages
        x_UIM_rel = x_UIM_authority*D/R
        x_PUM_rel = x_PUM_authority*D/R
        x_TST_rel = x_TST_authority*D/R
        x_ALL_rel = x_ALL_authority*D/R
        invCnorm = 1/C/R

        if temp_model_index == 1:
            if plot_selection == 'all' or plot_selection == 'olg':
                bp1 = BodePlot(title=f'{ifo} DARM Open Loop Gain: G', figsize=figsize)
                list_of_figures.append(plt.gcf())
            if plot_selection == 'all':
                bp2 = BodePlot(title=f'{ifo} DARM Open Loop Gain: G (ZOOM)', figsize=figsize)
                list_of_figures.append(plt.gcf())
            if plot_selection == 'all' or plot_selection == 'clg':
                bp3 = BodePlot(title=f'{ifo} DARM Closed Loop Gain: 1/(1+G)', figsize=figsize)
                list_of_figures.append(plt.gcf())
                bp3b = BodePlot(title=f'{ifo} DARM Closed Loop Gain: 1/(1+G) (ZOOM)',
                                figsize=figsize)
                list_of_figures.append(plt.gcf())
            if plot_selection == 'all' or plot_selection == 'digital':
                bp4 = BodePlot(title=f'{ifo} DARM Digital Filter: D', figsize=figsize)
                list_of_figures.append(plt.gcf())
            if plot_selection == 'all' or plot_selection == 'optical':
                bp5 = BodePlot(title=f'{ifo} Optical Plant Response: C', figsize=figsize)
                list_of_figures.append(plt.gcf())
            if plot_selection == 'all' or plot_selection == 'actuation':
                bp6 = BodePlot(title=f'{ifo} Act. stage strength', figsize=figsize)
                list_of_figures.append(plt.gcf())
                bp7 = BodePlot(title=f'{ifo} Act. stage digital filters', figsize=figsize)
                list_of_figures.append(plt.gcf())
                bp8 = BodePlot(title=f'{ifo} Act. stage LOCK_IN to Displ.', figsize=figsize)
                list_of_figures.append(plt.gcf())
                bp9 = BodePlot(title=f'{ifo} Act. stage LOCK_IN to Displ. (low zoom)',
                               figsize=figsize)
                list_of_figures.append(plt.gcf())
                bp10 = BodePlot(title=f'{ifo} Act. stage LOCK_IN to Displ. (high zoom)',
                                figsize=figsize)
                list_of_figures.append(plt.gcf())
                bp11 = BodePlot(title=f'{ifo} Act. stage DARM_IN to Disp. (D*A)', figsize=figsize)
                list_of_figures.append(plt.gcf())
                bp12 = BodePlot(title=f'{ifo} Act. stage relative contributions', figsize=figsize,
                                only='magnitude')
                list_of_figures.append(plt.gcf())

        ugf_freq, phase_margin = get_ugf(freq, G, search_start=ugf_start, search_end=ugf_end)
        phase_margin = 180+phase_margin

        if plot_selection == 'all' or plot_selection == 'olg':
            bp1.plot(freq, G, **kwargs)
            if label:
                bp1.legend(label)
            bp1.xlim(freq_min, freq_max)
            if greed:
                bp1.greed()
            bp1.vlines(ugf_freq)
            bp1.text(freq_min*1.1, 1.1*10**(-(temp_model_index)),
                     f'UGF {temp_model_index} = {ugf_freq}Hz, {phase_margin}deg')
            temp_bottom, temp_top = bp1.ax_mag.get_ylim()
            bp1.ylim(temp_top/default_olg_yspan, temp_top)

        if plot_selection == 'all':
            bp2.plot(freq, G, **kwargs)
            if label:
                bp2.legend(label)
            bp2.xlim(default_olg_zoom_fmin, default_olg_zoom_fmax)
            if greed:
                bp2.greed()
            bp2.vlines(ugf_freq)
            bp2.text(default_olg_zoom_fmin*1.04, 10**((1-temp_model_index)/8),
                     f'UGF {temp_model_index} = {ugf_freq}Hz, {phase_margin}deg')

        if plot_selection == 'all' or plot_selection == 'clg':
            bp3.plot(freq, 1./(1.+G), **kwargs)
            bp3b.plot(freq, 1./(1.+G), **kwargs)
            if label:
                bp3.legend(label)
                bp3b.legend(label)
            bp3.xlim(freq_min, freq_max)
            bp3b.xlim(default_clg_zoom_fmin, default_clg_zoom_fmax)
            if greed:
                bp3.greed()
                bp3b.greed()

        if plot_selection == 'all' or plot_selection == 'digital':
            bp4.plot(freq, D, **kwargs)
            if label:
                bp4.legend(label)
            bp4.xlim(freq_min, freq_max)
            if greed:
                bp4.greed()
            temp_bottom, temp_top = bp4.ax_mag.get_ylim()
            bp4.ylim(temp_top/default_dig_yspan, temp_top)

        if plot_selection == 'all' or plot_selection == 'optical':
            bp5.plot(freq, C, **kwargs)
            if label:
                bp5.legend(label)
            bp5.xlim(freq_min, freq_max)
            if greed:
                bp5.greed()

        if plot_selection == 'all' or plot_selection == 'actuation':
            if temp_model_index == 2:
                bp6.plot(freq, x_UIM_strength, ls='--')
                bp6.plot(freq, x_PUM_strength, ls='--')
                bp6.plot(freq, x_TST_strength, ls='--')
            else:
                bp6.plot(freq, x_UIM_strength, **kwargs)
                bp6.plot(freq, x_PUM_strength, **kwargs)
                bp6.plot(freq, x_TST_strength, **kwargs)
                bp6.ax_mag.set_prop_cycle(None)
                bp6.ax_phase.set_prop_cycle(None)
            bp6.xlim(freq_min, freq_max)
            temp_bottom, temp_top = bp6.ax_mag.get_ylim()
            bp6.ylim(temp_top/default_act_strength_yspan, temp_top)

            if greed:
                bp6.greed()

            if temp_model_index == 2:
                bp7.plot(freq, x_UIM_umodel, ls='--')
                bp7.plot(freq, x_PUM_umodel, ls='--')
                bp7.plot(freq, x_TST_umodel, ls='--')
            else:
                bp7.plot(freq, x_UIM_umodel, **kwargs)
                bp7.plot(freq, x_PUM_umodel, **kwargs)
                bp7.plot(freq, x_TST_umodel, **kwargs)
                bp7.ax_mag.set_prop_cycle(None)
                bp7.ax_phase.set_prop_cycle(None)
            bp7.xlim(freq_min, freq_max)

            if greed:
                bp7.greed()

            if temp_model_index == 2:
                bp8.plot(freq, x_UIM_authority, ls='--')
                bp8.plot(freq, x_PUM_authority, ls='--')
                bp8.plot(freq, x_TST_authority, ls='--')
                bp8.plot(freq, x_ALL_authority, ls='--')
            else:
                bp8.plot(freq, x_UIM_authority, **kwargs)
                bp8.plot(freq, x_PUM_authority, **kwargs)
                bp8.plot(freq, x_TST_authority, **kwargs)
                bp8.plot(freq, x_ALL_authority, **kwargs)
                bp8.ax_mag.set_prop_cycle(None)
                bp8.ax_phase.set_prop_cycle(None)
            bp8.xlim(freq_min, freq_max)
            temp_bottom, temp_top = bp8.ax_mag.get_ylim()
            bp8.ylim(temp_top/default_act_displ_yspan, temp_top)

            if greed:
                bp8.greed()

            if temp_model_index == 2:
                bp9.plot(freq, x_UIM_authority, ls='--')
                bp9.plot(freq, x_PUM_authority, ls='--')
                bp9.plot(freq, x_TST_authority, ls='--')
                bp9.plot(freq, x_ALL_authority, ls='--')
            else:
                bp9.plot(freq, x_UIM_authority, **kwargs)
                bp9.plot(freq, x_PUM_authority, **kwargs)
                bp9.plot(freq, x_TST_authority, **kwargs)
                bp9.plot(freq, x_ALL_authority, **kwargs)
                bp9.ax_mag.set_prop_cycle(None)
                bp9.ax_phase.set_prop_cycle(None)
            bp9.xlim(1e-1, 1e2)

            if greed:
                bp9.greed()

            if temp_model_index == 2:
                bp10.plot(freq, x_UIM_authority, ls='--')
                bp10.plot(freq, x_PUM_authority, ls='--')
                bp10.plot(freq, x_TST_authority, ls='--')
                bp10.plot(freq, x_ALL_authority, ls='--')
            else:
                bp10.plot(freq, x_UIM_authority, **kwargs)
                bp10.plot(freq, x_PUM_authority, **kwargs)
                bp10.plot(freq, x_TST_authority, **kwargs)
                bp10.plot(freq, x_ALL_authority, **kwargs)
                bp10.ax_mag.set_prop_cycle(None)
                bp10.ax_phase.set_prop_cycle(None)
            bp10.xlim(1e2, 1e3)

            if greed:
                bp10.greed()

            if temp_model_index == 2:
                bp11.plot(freq, x_UIM_total, ls='--')
                bp11.plot(freq, x_PUM_total, ls='--')
                bp11.plot(freq, x_TST_total, ls='--')
                bp11.plot(freq, (x_UIM_total+x_PUM_total+x_TST_total), ls='--')
            else:
                bp11.plot(freq, x_UIM_total, **kwargs)
                bp11.plot(freq, x_PUM_total, **kwargs)
                bp11.plot(freq, x_TST_total, **kwargs)
                bp11.plot(freq, (x_UIM_total+x_PUM_total+x_TST_total), **kwargs)
                bp11.ax_mag.set_prop_cycle(None)
                bp11.ax_phase.set_prop_cycle(None)
            bp11.xlim(freq_min, freq_max)
            temp_bottom, temp_top = bp11.ax_mag.get_ylim()
            bp11.ylim(temp_top/default_act_displ_yspan, temp_top)

            if greed:
                bp11.greed()

            if temp_model_index == 2:
                bp12.plot(freq, x_UIM_rel, ls='--')
                bp12.plot(freq, x_PUM_rel, ls='--')
                bp12.plot(freq, x_TST_rel, ls='--')
                bp12.plot(freq, x_ALL_rel, ls='--')
                bp12.plot(freq, invCnorm, ls='--')
            else:
                bp12.plot(freq, x_UIM_rel, **kwargs)
                bp12.plot(freq, x_PUM_rel, **kwargs)
                bp12.plot(freq, x_TST_rel, **kwargs)
                bp12.plot(freq, x_ALL_rel, **kwargs)
                bp12.plot(freq, invCnorm, **kwargs)
                bp12.ax_mag.set_prop_cycle(None)
                # bp12.ax_phase.set_prop_cycle(None)
            bp12.xlim(5, 1000)
            bp12.ylim(1e-3, 4e0)

            if greed:
                bp12.greed()

        if label:
            if plot_selection == 'all' or plot_selection == 'actuation':
                combined_model_label = []
                for labels in label:
                    combined_model_label.append(labels+' UIM')
                    combined_model_label.append(labels+' PUM')
                    combined_model_label.append(labels+' TST')
                bp6.legend(combined_model_label)
                bp7.legend(combined_model_label)
                combined_model_label_tot = []
                for labels in label:
                    combined_model_label_tot.append(labels+' UIM')
                    combined_model_label_tot.append(labels+' PUM')
                    combined_model_label_tot.append(labels+' TST')
                    combined_model_label_tot.append(labels+' Total')
                bp8.legend(combined_model_label_tot)
                bp9.legend(combined_model_label_tot)
                bp10.legend(combined_model_label_tot)
                bp11.legend(combined_model_label_tot)
                combined_model_label_tot = []
                for labels in label:
                    combined_model_label_tot.append(labels+' UIM*D/R')
                    combined_model_label_tot.append(labels+' PUM*D/R')
                    combined_model_label_tot.append(labels+' TST*D/R')
                    combined_model_label_tot.append(labels+' Total')
                    combined_model_label_tot.append(labels+' 1/C')
                bp12.legend(combined_model_label_tot)
        else:
            if plot_selection == 'all' or plot_selection == 'actuation':
                combined_model_label = []
                combined_model_label.append('UIM')
                combined_model_label.append('PUM')
                combined_model_label.append('TST')
                bp6.legend(combined_model_label)
                bp7.legend(combined_model_label)
                combined_model_label_tot = []
                combined_model_label_tot.append('UIM')
                combined_model_label_tot.append('PUM')
                combined_model_label_tot.append('TST')
                combined_model_label_tot.append('Total')
                bp8.legend(combined_model_label_tot)
                bp9.legend(combined_model_label_tot)
                bp10.legend(combined_model_label_tot)
                bp11.legend(combined_model_label_tot)
                combined_model_label_tot = []
                combined_model_label_tot.append('UIM*D/R')
                combined_model_label_tot.append('PUM*D/R')
                combined_model_label_tot.append('TST*D/R')
                combined_model_label_tot.append('Total*D/R')
                combined_model_label_tot.append('1/C')
                bp12.legend(combined_model_label_tot)
        temp_model_index = temp_model_index+1

    if filename:
        save_multifig(filename, list_of_figures)
    if show:
        plt.show()



import unittest
import pydarm
import numpy as np
import math
import h5py
import os
from pydarm.FIR import (FIRfilter, createFIRfilter,
                        correctFIRfilter, two_tap_zero_filter_response,
                        check_td_vs_fd, check_td_vs_fd_response)


class TestFIRfilter(unittest.TestCase):

    def setUp(self):
        self.advance_array = np.array([ 1.+0.00000000e+00j, -1.+1.22464680e-16j,  1.-2.44929360e-16j,
                                        -1.+3.67394040e-16j,  1.-4.89858720e-16j, -1.+6.12323400e-16j,
                                         1.-7.34788079e-16j, -1.+8.57252759e-16j,  1.-9.79717439e-16j])
        self.delay_array = np.array([ 1.+0.00000000e+00j, -1.-1.22464680e-16j,  1.+2.44929360e-16j,
                                     -1.-3.67394040e-16j,  1.+4.89858720e-16j, -1.-6.12323400e-16j,
                                      1.+7.34788079e-16j, -1.-8.57252759e-16j,  1.+9.79717439e-16j])
        self.delay_samples = 8
        self.df = 1.0
        self.dt = 0.0625
        self.freq_array = np.array([0., 1., 2., 3., 4., 5., 6., 7., 8.])
        self.latency = 0.5
        self.samples_to_HPcorner = 1
        self.samples_to_LPcorner = 0
        self.window = np.array([0.00233883, 0.02623098, 0.09449865, 0.22677684, 0.42285187,
                       0.65247867, 0.85980208, 0.98349415, 0.98349415, 0.85980208,
                       0.65247867, 0.42285187, 0.22677684, 0.09449865, 0.02623098,
                       0.00233883])

    def tearDown(self):
        del self.window
        del self.advance_array
        del self.delay_array
        del self.delay_samples
        del self.df
        del self.dt
        del self.freq_array
        del self.latency
        del self.samples_to_HPcorner
        del self.samples_to_LPcorner

    def test_FIRfilter(self):
        test_FIRfilter = FIRfilter(fNyq=8, dur=1, highpass_fcut=4, lowpass_fcut=None,
                                latency=None, window_type='kaiser', freq_res=3.0)

        self.assertEqual(test_FIRfilter.delay_samples, self.delay_samples)
        self.assertEqual(test_FIRfilter.df, self.df)
        self.assertEqual(test_FIRfilter.dt, self.dt)
        self.assertEqual(test_FIRfilter.latency, self.latency)
        self.assertEqual(test_FIRfilter.samples_to_HPcorner, self.samples_to_HPcorner)
        self.assertEqual(test_FIRfilter.samples_to_LPcorner, self.samples_to_LPcorner)
        self.assertEqual(len(test_FIRfilter.advance_array), len(self.advance_array))
        for n in range(len(self.advance_array)):
            self.assertAlmostEqual(np.abs(test_FIRfilter.advance_array[n]) / np.abs(self.advance_array[n]), 1.0)
            self.assertAlmostEqual(np.angle(test_FIRfilter.advance_array[n], deg=True) - np.angle(self.advance_array[n], deg=True), 0.0)
        self.assertEqual(len(test_FIRfilter.delay_array), len(self.delay_array))
        for n in range(len(self.delay_array)):
            self.assertAlmostEqual(np.abs(test_FIRfilter.delay_array[n]) / np.abs(self.delay_array[n]), 1.0)
            self.assertAlmostEqual(np.angle(test_FIRfilter.delay_array[n], deg=True) - np.angle(self.delay_array[n], deg=True), 0.0)
        self.assertEqual(len(test_FIRfilter.freq_array), len(self.freq_array))
        for n in range(len(self.freq_array)):
            self.assertAlmostEqual(test_FIRfilter.freq_array[n], self.freq_array[n])
        self.assertEqual(len(test_FIRfilter.window), len(self.window))
        for n in range(len(self.window)):
            self.assertAlmostEqual(test_FIRfilter.window[n] / self.window[n], 1.0, places=5)


class TestCreateFIRfilter(unittest.TestCase):

    def setUp(self):
        self.FIRPars = FIRfilter(fNyq=8, dur=1, highpass_fcut=4, lowpass_fcut=None,
                                 latency=None, window_type='kaiser', freq_res=3.0)
        self.config = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
whitening_mode_names = mode1, mode1
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
omc_meas_p_whitening_uncompensated_mode1   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules =
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_mode1 = ON, ON
'''
        self.known_Cfir = np.array([-6.63850419e+01,  8.93362543e+02,  8.56077110e+04, -2.27887893e+05,
                                    -8.86759535e+05,  1.61591927e+06,  1.23433242e+06, -2.69276290e+06,
                                     1.79904168e+06, -1.74942704e+06,  6.65554086e+05,  9.35475644e+05,
                                    -4.10502392e+05, -8.87882831e+04,  2.10864177e+04,  1.42418753e+02])

    def tearDown(self):
        del self.FIRPars
        del self.config
        del self.known_Cfir

    def test_createFIRfilter(self):
        C = pydarm.sensing.SensingModel(self.config)
        Cf = C.compute_sensing(self.FIRPars.freq_array)
        test_Cfir, test_Cmodel = createFIRfilter(self.FIRPars, Cf)

        for n in range(len(self.known_Cfir)):
            self.assertAlmostEqual(test_Cfir[n] /
                                   self.known_Cfir[n],
                                   1.0)

class TestCorrectFIRfilter(unittest.TestCase):

    def setUp(self):
        self.FIRPars = FIRfilter(fNyq=8, dur=1, highpass_fcut=4, lowpass_fcut=None,
                                 latency=None, window_type='kaiser', freq_res=3.0)
        self.config = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = mode1, mode1
omc_meas_p_whitening_uncompensated_mode1   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules =
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_mode1 = ON, ON
'''

        self.known_Corfir = np.array([-169801.88070367-2.19120483e+02j,
                               -698906.37183737-2.74819393e+04j,
                              -2030852.42623273+8.40205385e+03j,
                             -54502988.18785564+4.75790494e+07j,
                              47217941.0142496 +8.94536006e+06j,
                               6654278.33719367-1.81089469e+05j,
                               4974433.40023975-4.73425087e+05j,
                               4662225.0011754 -2.41147877e+05j])

    def tearDown(self):
        del self.FIRPars
        del self.config
        del self.known_Corfir

    def test_correctFIRfilter(self):
        C = pydarm.sensing.SensingModel(self.config)
        Cf = C.compute_sensing(self.FIRPars.freq_array)
        test_Cfir = createFIRfilter(self.FIRPars, Cf)[0]
        test_Corfir = correctFIRfilter(self.FIRPars, test_Cfir, Cf,  [2, 4, 7, 9])
        # The first element is zero so we will delete this one from the check
        # to avoid divide by zero errors
        test_Corfir = np.delete(test_Corfir, 0)

        self.assertEqual(len(test_Corfir), len(self.known_Corfir))
        for n in range(len(test_Corfir)):
            self.assertAlmostEqual(
                np.abs(test_Corfir[n])/np.abs(self.known_Corfir[n]), 1)
            self.assertAlmostEqual(
                np.angle(test_Corfir[n], deg=True) -
                np.angle(self.known_Corfir[n], deg=True), 0)


class Testtwo_tap_zero_filter(unittest.TestCase):

    def setUp(self):
        self.known_two_tap_zero_filt = np.array(
            [0.999999999999999889+0.000000000000000000j,
             0.944065144524670274-0.242141032019198316j,
             0.791790257603250391-0.430462197742083885j,
             0.584088546831189381-0.528606912083167346j,
             0.372528221948197957-0.528347694598477635j,
             0.201488911932644205-0.449462235973869528j,
             0.094304975193611695-0.329261396976028187j,
             0.048712779360778842-0.206416348868232596j,
             0.042948901676579269-0.105952738743734842j,
             0.049370136034108433-0.031598975501051528j])

    def tearDown(self):
        del self.known_two_tap_zero_filt

    def test_two_tap_zero_filt(self):
        test_ttzf = two_tap_zero_filter_response([1, 2], 1, np.linspace(1, 100, 10))

        for n in range(len(test_ttzf)):
            self.assertAlmostEqual(np.abs(test_ttzf[n]),
                                   np.abs(self.known_two_tap_zero_filt[n]), places=6)
            self.assertAlmostEqual(np.angle(test_ttzf[n], deg=True),
                                   np.angle(self.known_two_tap_zero_filt[n], deg=True),
                                   places=6)


class Testcheck_td_vs_fd(unittest.TestCase):

    def setUp(self):
        self.FIRPars = FIRfilter(fNyq=2048, dur=1, highpass_fcut=8, lowpass_fcut=None,
                                 latency=None, window_type='kaiser', freq_res=3.0)
        self.config = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
whitening_mode_names = mode1, mode1
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
omc_meas_p_whitening_uncompensated_mode1   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules =
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_mode1 = ON, ON
'''

    def tearDown(self):
        del self.FIRPars
        del self.config

    def test_check_td_vs_fd(self):
        C = pydarm.sensing.SensingModel(self.config)
        Cf = C.compute_sensing(self.FIRPars.freq_array)
        test_Cfir = createFIRfilter(self.FIRPars, Cf)[0]
        ctvf = check_td_vs_fd(test_Cfir, Cf, fNyq=self.FIRPars.fNyquist,
                                   delay_samples=self.FIRPars.delay_samples)

        test_freq = ctvf[0][0]
        test_mag_ratios = ctvf[1][0]
        test_phase_diffs = ctvf[2][0]

        df = test_freq[1] - test_freq[0]
        # For this configuration, check that the magnitude and phase are close to
        # 1 and 0, respectively between 50 and 1000 Hz
        Nmin = math.ceil(40 / df)
        Nmax = math.ceil(1000 / df)
 
        for n in range(Nmin, Nmax):
            self.assertLess(np.abs(test_mag_ratios[n]), 1.01)
            self.assertLess(np.abs(test_phase_diffs[n]), 1)


class TestCheckTdVsFdResponse(unittest.TestCase):

    def setUp(self):
        self.FIRPars = FIRfilter(fNyq=2048, dur=1, highpass_fcut=8, lowpass_fcut=None,
                                 latency=None, window_type='kaiser', freq_res=3.0)
        self.config = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
whitening_mode_names = mode1, mode1
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
omc_meas_p_whitening_uncompensated_mode1   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules =
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_mode1 = ON, ON

[digital]
digital_filter_file    = test/H1OMC_1239468752.txt
digital_filter_bank    = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain    = 400,1

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x    = OFF, ON, ON, ON
darm_feedback_y    = OFF, OFF, OFF, OFF

[actuation_x_arm]
darm_feedback_sign = -1
uim_NpA       = 1.634
pum_NpA       = 0.02947
tst_NpV2      = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank    = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain    = 1.0
tst_lock_bank       = ETMX_L3_LOCK_L
tst_lock_modules    = 5,8,9,10
tst_lock_gain       = 1.0
tst_drive_align_bank     = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules  = 4,5
tst_drive_align_gain     = -35.7
pum_lock_bank    = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain    = 23.0
pum_drive_align_bank    = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain    = 1.0
pum_coil_outf_signflip  = 1
uim_lock_bank    = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain    = 1.06
uim_drive_align_bank    = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain    = 1.0
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
pum_driver_DC_trans_ApV = 2.6847e-4
uim_driver_DC_trans_ApV = 6.1535e-4
anti_imaging_rate_string = 16k
anti_imaging_method      = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
uim_delay = 0
pum_delay = 0
tst_delay = 0

[pcal]
pcal_filter_file           = test/H1CALEY_1123041152.txt
pcal_filter_bank           = PCALY_TX_PD
pcal_filter_modules_in_use = 6,8
pcal_filter_gain           = 1.0
pcal_dewhiten               = 1.0, 1.0
pcal_incidence_angle        = 8.8851
pcal_etm_watts_per_ofs_volt = 0.13535
ref_pcal_2_darm_act_sign    = -1.0
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat
'''
    def tearDown(self):
        del self.FIRPars
        del self.config

    def test_check_td_vs_fd_response(self):
        freq = self.FIRPars.freq_array
        # DARM model throws warnings when computed at 0 Hz, so remove that from frequency array
        freq = np.delete(freq, 0)
        darm = pydarm.darm.DARMModel(self.config)
        # To get frequency vectors correct, need to reinsert the zero frequency component after DARM components are computed
        # Just set this DC component to 0
        InvC = np.insert(1/darm.sensing.compute_sensing(freq), 0, 0)
        TST = np.insert(darm.actuation.xarm.compute_actuation_single_stage(freq, stage='TST'), 0, 0)
        PUM = np.insert(darm.actuation.xarm.compute_actuation_single_stage(freq, stage='PUM'), 0, 0)
        UIM = np.insert(darm.actuation.xarm.compute_actuation_single_stage(freq, stage='UIM'), 0, 0)
        D = np.insert(darm.digital.compute_response(freq), 0, 0)
        R = np.insert(darm.compute_response_function(freq), 0, 0)

        InvC_filt = createFIRfilter(self.FIRPars, InvC)[0]
        TST_filt = createFIRfilter(self.FIRPars, TST)[0]
        PUM_filt = createFIRfilter(self.FIRPars, PUM)[0]
        UIM_filt = createFIRfilter(self.FIRPars, UIM)[0]

        test_freq, test_ratio_mag, test_ratio_pha = check_td_vs_fd_response(InvC_filt,
                                            None,
                                            TST_filt,
                                            PUM_filt,
                                            UIM_filt,
                                            None,
                                            D,
                                            R,
                                            time_delay = 0.0,
                                            invsens_fNyq=self.FIRPars.fNyquist,
                                            act_fNyq=self.FIRPars.fNyquist,
                                            D_fNyq=self.FIRPars.fNyquist,
                                            R_fNyq=self.FIRPars.fNyquist)

        
        df = test_freq[1] - test_freq[0]
        # For this configuration, check that the magnitude and phase are close to
        # 1 and 0, respectively between 50 and 1000 Hz
        Nmin = math.ceil(40 / df)
        Nmax = math.ceil(1000 / df)
        
        for n in range(Nmin, Nmax):
            self.assertLess(np.abs(test_ratio_mag[n]), 1.05)
            self.assertLess(np.abs(test_ratio_pha[n]), 5)

class TestGDS_FIR_filter_generation(unittest.TestCase):
    def setUp(self):
        # Maddie test
        # Set up for control chain FIR filter generation
        self.FIRpars = FIRfilter(fNyq=1024, dur=3.5, highpass_fcut=10.5, lowpass_fcut=None,
                                 latency=None, window_type='dpss', freq_res=4.0)

        # Load in known transfer function and resulting FIR filter
        h5f = h5py.File('./test/FIR_unit_test_coeffs.h5', 'r')
        self.known_FIR_filter = h5f['FIR_filter'][:]
        self.known_tf = h5f['transfer_function'][:]

    def tearDown(self):
        del self.FIRpars
        del self.known_FIR_filter
        del self.known_tf

    def test_GDS_FIR_filter_generation(self):
        # Generate test FIR filter from frequency domain transfer function
        [test_FIR_filter, model] = createFIRfilter(self.FIRpars, self.known_tf)
        # FIXME: (Arif) Scipy and FIRtools have much different results under 10 Hz.
        # I changed the range and places. My local test could take higher places.
        for n in range(300, len(self.known_FIR_filter)-300):
            self.assertAlmostEqual(abs((self.known_FIR_filter[n] / test_FIR_filter[n])
                                       - 1), 0, places=3)


class TestFilterGeneration(unittest.TestCase):

    def setUp(self):
        self.arm_length = 3994.4698
        self.fcc = 410.6
        self.fs = 4.468
        self.fs_squared = 19.963024
        self.srcQ = 52.14
        self.ips = 1.0
        os.environ['CAL_DATA_ROOT'] = './test'

    def tearDown(self):
        del self.arm_length
        del self.fcc
        del self.fs
        del self.fs_squared
        del self.srcQ
        del self.ips
        del os.environ['CAL_DATA_ROOT']

    def test_FilterGeneration(self):
        config = './example_model_files/H1_20190416.ini'
        FG = pydarm.FIR.FilterGeneration(config)
        self.assertEqual(self.arm_length, FG.arm_length)
        self.assertEqual(self.fcc, FG.fcc)
        self.assertEqual(self.fs, FG.fs)
        self.assertEqual(self.fs_squared, FG.fs_squared)
        self.assertEqual(self.srcQ, FG.srcQ)
        self.assertEqual(self.ips, FG.ips)


class TestGDS(unittest.TestCase):

    def setUp(self):
        self.GDS_file = h5py.File('./test/GDS_test.h5', 'r')
        self.ctrl_corr_td = self.GDS_file['ctrl_corr_filter'][:]
        os.environ['CAL_DATA_ROOT'] = './test'

    def tearDown(self):
        del self.GDS_file
        del self.ctrl_corr_td
        del os.environ['CAL_DATA_ROOT']

    def test_GDS(self):
        config = './example_model_files/H1_20190416.ini'
        FG = pydarm.FIR.FilterGeneration(config)
        FG.GDS(ctrl_window_type='kaiser', res_window_type='kaiser',
               make_plot=False, output_filename='./test/GDS.npz',
               plots_directory='./examples/GDS_plots')
        gds = np.load('./test/GDS.npz')
        ctrl_td = gds['ctrl_corr_filter']
        for n in range(len(self.ctrl_corr_td)):
            self.assertAlmostEqual(abs((self.ctrl_corr_td[n] / ctrl_td[n])
                                       - 1), 0, places=3)


class TestDCS(unittest.TestCase):

    def setUp(self):
        self.DCS_file = h5py.File('./test/DCS_test.h5', 'r')
        self.known_act_tst = self.DCS_file['actuation_tst'][:]
        os.environ['CAL_DATA_ROOT'] = './test'

    def tearDown(self):
        del self.DCS_file
        del self.known_act_tst
        del os.environ['CAL_DATA_ROOT']

    def test_DCS(self):
        config = './example_model_files/H1_20190416.ini'
        FG = pydarm.FIR.FilterGeneration(config)
        FG.DCS(act_window_type='kaiser', invsens_window_type='kaiser',
               make_plot=False, output_filename='./test/DCS.npz',
               plots_directory='examples/DCS_plots')
        dcs = np.load('./test/DCS.npz')
        act_tst = dcs['actuation_tst']
        for n in range(len(self.known_act_tst)):
            self.assertAlmostEqual(abs((self.known_act_tst[n] / act_tst[n])
                                       - 1), 0, places=3)


if __name__ == '__main__':
    unittest.main()



import unittest
import pydarm
import numpy as np


class TestFindPrimeFactors(unittest.TestCase):

    def setUp(self):
        self.N = 3
        self.known_results = np.array([3, 1])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_find_prime_factors(self):
        test = pydarm.FIRtools.find_prime_factors(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.real(test[n]),
                                   np.real(self.known_results[n]))


class TestFindM(unittest.TestCase):

    def setUp(self):
        self.N = 10
        self.known_results_0 = 10
        self.known_results_1 = np.array([2, 5, 1])

    def tearDown(self):
        del self.N
        del self.known_results_0
        del self.known_results_1

    def test_find_M(self):
        test = pydarm.FIRtools.find_M(self.N)

        self.assertEqual(test[0], self.known_results_0)
        for n in range(len(test[1])):
            self.assertAlmostEqual(np.real(test[1][n]),
                                   np.real(self.known_results_1[n]))


class TestFindExpArray(unittest.TestCase):

    def setUp(self):
        self.N = 10
        self.known_results = np.array([
            1.+0.j,
            0.80901699-0.58778525j,
            0.30901699-0.95105652j,
            -0.30901699-0.95105652j,
            -0.80901699-0.58778525j,
            -1.+0.j,
            -0.80901699+0.58778525j,
            -0.30901699+0.95105652j,
            0.30901699+0.95105652j,
            0.80901699+0.58778525j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_find_exp_array(self):
        test = pydarm.FIRtools.find_exp_array(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestFindExpArray2(unittest.TestCase):

    def setUp(self):
        self.N = 10
        self.known_results = np.array(
            [1.+0.j,
             0.95105652-0.30901699j,
             0.30901699-0.95105652j,
             -0.95105652-0.30901699j,
             0.30901699+0.95105652j,
             0.-1.j,
             0.30901699+0.95105652j,
             -0.95105652-0.30901699j,
             0.30901699-0.95105652j,
             0.95105652-0.30901699j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_find_exp_array2(self):
        test = pydarm.FIRtools.find_exp_array2(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestDFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [55.+0.j,
             -5.+15.38841769j,
             -5.+6.8819096j,
             -5.+3.63271264j,
             -5.+1.62459848j,
             -5.+0.j,
             -5.-1.62459848j,
             -5.-3.63271264j,
             -5.-6.8819096j,
             -5.-15.38841769j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_dft(self):
        test = pydarm.FIRtools.dft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestRDFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [55.+0.j,
             -5.+15.38841769j,
             -5.+6.8819096j,
             -5.+3.63271264j,
             -5.+1.62459848j,
             -5.+0.j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_rdft(self):
        test = pydarm.FIRtools.rdft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestIRDFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [5.50000000e+00,
             -1.84241319e+00,
             -1.80700362e-20,
             -2.22222222e-01,
             -7.83034902e-20,
             -9.46715662e-02,
             0.00000000e+00,
             -6.29152406e-02,
             -7.83034902e-20,
             -5.55555556e-02,
             -7.83034902e-20,
             -6.29152406e-02,
             0.00000000e+00,
             -9.46715662e-02,
             -7.83034902e-20,
             -2.22222222e-01,
             -1.80700362e-20,
             -1.84241319e+00])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_irdft(self):
        test = pydarm.FIRtools.irdft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))


class TestFFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [55.+0.j,
             -5.+15.38841769j,
             -5.+6.8819096j,
             -5.+3.63271264j,
             -5.+1.62459848j,
             -5.+0.j,
             -5.-1.62459848j,
             -5.-3.63271264j,
             -5.-6.8819096j,
             -5.-15.38841769j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_fft(self):
        test = pydarm.FIRtools.fft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestIFFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [5.5+0.j,
             -0.5-1.538841768587627j,
             -0.5-0.688190960235587j,
             -0.5-0.36327126400268j,
             -0.5-0.162459848116453j,
             -0.5+0.j,
             -0.5+0.162459848116453j,
             -0.5+0.36327126400268j,
             -0.5+0.688190960235587j,
             -0.5+1.538841768587627j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_ifft(self):
        test = pydarm.FIRtools.ifft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestPrime_FFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [55.+2.602085213965211e-19j,
             -5.+1.538841768587627e+01j,
             -5.+6.881909602355868e+00j,
             -5.+3.632712640026804e+00j,
             -5.+1.624598481164532e+00j,
             -5.-3.079134169858833e-18j,
             -5.-1.624598481164532e+00j,
             -5.-3.632712640026804e+00j,
             -5.-6.881909602355868e+00j,
             -5.-1.538841768587627e+01j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_prime_fft(self):
        test = pydarm.FIRtools.prime_fft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestPrime_IRFFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [5.500000000000000e+00, -1.842413193195909e+00,
             4.549226020051922e-20, -2.222222222222222e-01,
             -8.417075210445289e-20, -9.467156616899152e-02,
             -1.070816960479511e-19, -6.291524063509968e-02,
             -1.051106487897057e-19, -5.555555555555556e-02,
             3.259193740688347e-20, -6.291524063509968e-02,
             -3.212450881438532e-20, -9.467156616899152e-02,
             -5.489919065095004e-20, -2.222222222222222e-01,
             2.254883592741939e-19, -1.842413193195909e+00])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_prime_irfft(self):
        test = pydarm.FIRtools.prime_irfft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))


class TestPrime_RFFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [55.-7.401486830834377e-18j,
             -5.+1.538841768587627e+01j,
             -5.+6.881909602355868e+00j,
             -5.+3.632712640026804e+00j,
             -5.+1.624598481164532e+00j,
             -5.+1.850371707708594e-18j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_prime_rfft(self):
        test = pydarm.FIRtools.prime_rfft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestRFFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [55.+0.j,
             -5.+15.388417685876267j,
             -5.+6.881909602355868j,
             -5.+3.632712640026804j,
             -5.+1.624598481164532j,
             -5.+0.j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_rfft(self):
        test = pydarm.FIRtools.rfft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class TestIRFFT(unittest.TestCase):

    def setUp(self):
        self.N = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.known_results = np.array(
            [5.500000000e+00, -1.842413193e+00, 6.023345403e-21,
             -2.222222222e-01, 1.204669081e-20, -9.467156617e-02,
             1.204669081e-20, -6.291524064e-02, 0.000000000e+00,
             -5.555555556e-02, 0.000000000e+00, -6.291524064e-02,
             1.204669081e-20, -9.467156617e-02, 1.204669081e-20,
             -2.222222222e-01, 9.035018104e-21, -1.842413193e+00])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_irfft(self):
        test = pydarm.FIRtools.irfft(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(
                np.abs(np.angle(test[n], deg=True) -
                       np.angle(self.known_results[n], deg=True)), 0.0, places=6)


class Testmattimesvec(unittest.TestCase):

    def setUp(self):
        self.M = [1., 2., 3.]
        self.V = [1., 2.]
        self.known_results = [1.33333333333333, 1.]

    def tearDown(self):
        del self.M
        del self.V
        del self.known_results

    def test_mattimesvec(self):
        test = pydarm.FIRtools.mat_times_vec(self.M, self.V)
        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))


class TestDPSS(unittest.TestCase):

    def setUp(self):
        self.N = 5
        self.B = 16
        self.known_results = np.array(
            [0.44321078700654, 0.83098434177807, 1.,
             0.83098434177807, 0.44321078700654])

    def tearDown(self):
        del self.N
        del self.B
        del self.known_results

    def test_DPSS(self):
        test = pydarm.FIRtools.DPSS(self.N, self.B)
        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]),
                                   places=6)


class TestcomputeTn(unittest.TestCase):

    def setUp(self):
        self.N = 1
        self.B = 16
        self.known_results = 1.0

    def tearDown(self):
        del self.N
        del self.B
        del self.known_results

    def test_computeTn(self):
        test = pydarm.FIRtools.compute_Tn(self.N, self.B)
        self.assertAlmostEqual(np.abs(test),
                               np.abs(self.known_results))


class TestcomputeW0Lagged(unittest.TestCase):

    def setUp(self):
        self.N = 16
        self.B = 1
        self.known_results = [1.000000000e+00+0.j,
                              -1.065235759e-01-0.021188857j,
                              3.652655927e-02+0.015129796j,
                              4.151055100e-02+0.027736463j,
                              3.282099121e-02+0.032820991j,
                              2.101570991e-02+0.031452233j,
                              1.014723686e-02+0.024497597j,
                              2.659050841e-03+0.013367951j,
                              9.183429072e-20+0.j]

    def tearDown(self):
        del self.N
        del self.B
        del self.known_results

    def test_compute_W0_lagged(self):
        test = pydarm.FIRtools.compute_W0_lagged(self.N, self.B)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]))
            self.assertAlmostEqual(np.angle(test[n], deg=True),
                                   np.angle(self.known_results[n], deg=True),
                                   places=6)


class TestDolphChebyshev(unittest.TestCase):

    def setUp(self):
        self.N = 16
        self.B = 1
        self.known_results = np.array(
            [0.8668296927388, 0.50431006465412, 0.62167043818165,
             0.73337937658832, 0.8327331038598, 0.91351461128685,
             0.9705186581255,  1., 1., 0.9705186581255,
             0.91351461128685, 0.8327331038598, 0.73337937658832,
             0.62167043818165, 0.50431006465412, 0.8668296927388])

    def tearDown(self):
        del self.N
        del self.B
        del self.known_results

    def test_DolphChebyshev(self):
        test = pydarm.FIRtools.DolphChebyshev(self.N, self.B)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]), places=6)
            self.assertAlmostEqual(np.angle(test[n], deg=True),
                                   np.angle(self.known_results[n], deg=True),
                                   places=6)


class TestResample(unittest.TestCase):

    def setUp(self):
        self.N = np.linspace(1, 100, 15)
        self.B = 10
        self.known_results = np.array(
            [1., 12.01200936894652, 23.05579571636385,
             34.0362625611402, 45.1101331908868, 55.8898668091132,
             66.9637374388598, 77.94420428363615, 88.98799063105348,
             100.])

    def tearDown(self):
        del self.N
        del self.B
        del self.known_results

    def test_resample(self):
        test = pydarm.FIRtools.resample(self.N, self.B)
        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]), places=6)


class TestFreqresp(unittest.TestCase):

    def setUp(self):
        self.N = np.linspace(1, 16, 8)
        self.known_results = np.array(
            [68.+0.j,
             59.48445890636821-30.52945786235738j,
             36.84293730912102-51.76597925799235j,
             7.68035907617445-57.79361930967018j,
             -18.57943825259448-48.11882085320455j,
             -34.18355417297142-27.60239500215145j,
             -35.76197233728974-4.36222766254847j,
             -25.22421250295184+13.45302599634338j,
             -8.57142857142857+20.6932591060551j,
             6.78944776685062+16.84735056355831j,
             15.06065526360838+5.72520046505279j,
             14.15621988372335-6.55486359389886j,
             6.1001834190962-14.32465515694111j,
             -4.31159130661836-14.67005982624531j,
             -11.88468421883575-8.34956381179203j,
             -13.33966008951263+0.9510851460444j,
             -8.57142857142857+8.57142857142857j,
             -0.38653393620457+11.0850464457693j,
             7.04820384659785+7.76456399152353j,
             10.25421289199872+0.6991807859211j,
             8.02164711553033-6.395424105799j,
             1.82671829025581-10.04394696175532j,
             -5.03754860172615-8.64991512808857j,
             -9.14029290683677-3.15358857555265j,
             -8.57142857142857+3.55040196319796j,
             -3.79633017320841+8.10339541063871j,
             2.66677748140826+8.33474493830941j,
             7.54930744558597+4.2799094461385j,
             8.45760771796795-1.90387551634815j,
             5.02832789185148-7.05299088135736j,
             -0.93436874288388-8.57976326067755j,
             -6.3868770645046-5.74418800922619j,
             -8.57142857142857+0.j])

    def tearDown(self):
        del self.N
        del self.known_results

    def test_freqresp(self):
        test = pydarm.FIRtools.freqresp(self.N)

        for n in range(len(test)):
            self.assertAlmostEqual(np.abs(test[n]),
                                   np.abs(self.known_results[n]), places=6)
            self.assertAlmostEqual(np.angle(test[n], deg=True),
                                   np.angle(self.known_results[n], deg=True),
                                   places=6)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) Evan Goetz (2022)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy import signal

from .utils import load_foton_export_tf, digital_delay_filter
from .digital import daqdownsamplingfilters
from .model import Model
from .darm import DARMModel


class CALCSModel(DARMModel):

    def __init__(self, config, calcs=None):
        """
        Initialize a CALCSModel object

        Note that any string or path to file string in `calcs` will
        overwrite anything in the `config` parameter string or path to file
        """
        super().__init__(config)
        if 'calcs' in self._config:
            self.calcs = Model(config, measurement='calcs')
        if calcs is not None:
            self.calcs = Model(calcs, measurement='calcs')
        if not hasattr(self, 'calcs'):
            raise ValueError('No CALCS parameters have been defined')

    def optical_response_ratio(self, frequencies):
        """
        This computes (opt resp) / (opt resp)_foton

        It is a bit confusing because the FOTON filter is the inverse
        sensing function, so we'll be multiplying the true optical
        response from the model by the FOTON transfer function.

        In T1900169, in the definition of C/C_CALCS, the last three
        terms are
        .. math :: (opt resp)/(opt resp)_foton * (LP_foton/LP) * LP
        and this method returns this combination since the export from
        FOTON gives the full combination of
        .. math :: LP_foton / (opt resp)_foton

        Note that (LP_foton/LP) is the "IIR warping"

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the optical response ratio

        Returns
        -------
        foton_inv_sensing_interp : `complex128`, array-like
            ratio of interpolated values
        """

        # We need the coupled cavity LTI object
        coupled_cavity = self.sensing.optical_response(
            self.sensing.coupled_cavity_pole_frequency,
            self.sensing.detuned_spring_frequency,
            self.sensing.detuned_spring_q,
            pro_spring=self.sensing.is_pro_spring)

        # load inverse sensing data from foton file export (1/SRC_D2N)
        foton_freq, foton_tf = load_foton_export_tf(
            self.dpath(self.calcs.foton_invsensing_tf))

        # apply gain, default is 1.0
        invsensing_gain = getattr(self, 'invsensing_gain', 1.0)
        foton_tf *= invsensing_gain

        # Take ratio of true optical response / foton response where
        # "foton response" is already the approximated
        # LP_foton / (opt resp)_foton, so we just need to multiply the two
        # terms
        opt_response_ratio = (
            signal.freqresp(coupled_cavity, 2.0*np.pi*foton_freq)[1] *
            foton_tf)

        # interpolate to the requested frequencies
        opt_response_ratio_interp = np.interp(frequencies,
                                              foton_freq,
                                              opt_response_ratio)

        return opt_response_ratio_interp

    def sus_response_ratio(self, frequencies, arm, stage):
        """
        This computes (sus resp) / (sus resp)_foton

        See T1900169 definition of delta A_mn

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the optical response ratio
        arm : str, `x` or `y`
        stage : str `UIM`, `PUM`, or `TST`

        Returns
        -------
        response_ratio : `complex128`, array-like
            ratio of interpolated values
        """

        # load FOTON suspension dynamics filter
        foton_freq, foton_tf = load_foton_export_tf(
            self.dpath(getattr(self.calcs,
                               f'{arm.lower()}arm_{stage.lower()}_analog')))

        if arm.lower() == 'x':
            arm_act = self.actuation.xarm
        elif arm.lower() == 'y':
            arm_act = self.actuation.yarm
        else:
            raise ValueError('The accepted values for arm is x or y')

        # get true suspension dynamics
        [uim, pum, tst] = arm_act.matlab_force2length_response(
            arm_act.suspension_file, foton_freq)

        if stage.lower() == 'uim':
            ratio = uim / foton_tf
        elif stage.lower() == 'pum':
            ratio = pum / foton_tf
        elif stage.lower() == 'tst':
            ratio = tst / foton_tf
        else:
            raise ValueError('stage must be uim, pum or tst')

        # interpolate to the requested frequencies
        response_ratio = np.interp(frequencies, foton_freq, ratio)

        return response_ratio

    def C_corr(self, frequencies):
        """
        Compute delta C = delay*(opt resp)/(opt resp)_foton*C_r

        The inverse of this transfer function is multiplied by
        DELTAL_RESIDUAL as the GDS correction.

        We need to divide out the normalized optical response FOTON filter
        only and be sure to include 1 16k clock cycle delay from the model
        jump OMC to CAL-CS. There is an optical
        gain factor in the front end model, but it is fine to leave this in
        place. The output is basically C_pydarm / C_foton, and GDS will need to
        either divide the DELTAL_RESIDUAL data by this function or invert the output
        of this function and multiply by the DELTAL_RESIDUAL data.

        This is delay*(opt resp)/(opt resp)_foton*C_r (see T1900169-v5 eq 33)

        This is also known as C_corr or delta C in T1900169

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the GDS sensing correction

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the correction
        """

        # Residual sensing function transfer function
        # This is everything in sensing except the optical gain and response
        C_res = self.sensing.sensing_residual(frequencies)

        # we will need to apply a delay because when it is inverted in GDS
        # this will become an advance (the correct thing to do)
        omc_to_calcs_response = (
            signal.dfreqresp(digital_delay_filter(1, 2**14),
                             2.0*np.pi*frequencies/2**14)[1])

        # get the optical response ratio between FOTON and pyDARM
        opt_response_ratio = self.optical_response_ratio(frequencies)

        # the final correction is the sensing model divided by the interpolated
        # foton filter
        correction = (omc_to_calcs_response *
                      opt_response_ratio *
                      C_res)

        return correction

    def drivealign_out_to_longitudinal_displacement(
            self, frequencies, arm, stage):
        """
        CALCS representation of the transfer function from the
        DRIVEALIGN bank to longitudinal displacement of the test
        mass.
        Units are m/ct = (N/ct) * (m/N)


        The gain (N/ct) of the CALCS model is assumed to be exactly the same
        as what is in the reference model, so that is used. If any digital
        gain is present as <arm>_<stage>_analog_gain, then it will be
        applied here.

        If there is anything in the "coiloutf" key-value pair, then this
        transfer function is also used, but this would be used only in
        special circumstances.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the CALCS approximated longitudinal
            actuation of the test mass from the given stage
        arm : `str`
            arm for the calculation, 'x' or 'y'
        stage : `str`
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'

        Returns
        -------
        out_tf : `complex128`, array-like
            CALCS approximation of the DRIVEALIGN output to displacement
        """

        # The pieces of the actuator are multiplicative to obtain the final
        # result so we need to start with unity transfer functions and
        # change them only if a user has supplied a FOTON transfer
        # function export. The tf_ana_interp will be filled in by the FOTON
        # export of the SUS response. The tf_dig_interp will only be used if
        # there is CALCS COILOUTF content.
        tf_dig_interp = np.ones(len(frequencies), dtype='complex128')
        tf_ana_interp = np.ones(len(frequencies), dtype='complex128')

        # Get the coiloutf and analog exports if they exist
        if hasattr(self.calcs, f'{arm.lower()}arm_{stage.lower()}_coiloutf'):
            val = getattr(
                self.calcs, f'{arm.lower()}arm_{stage.lower()}_coiloutf')
            if val != '':
                [f, tf_dig] = load_foton_export_tf(self.dpath(val))
                tf_dig_interp = np.interp(frequencies, f, tf_dig)
        if hasattr(self.calcs, f'{arm.lower()}arm_{stage.lower()}_analog'):
            val = getattr(
                self.calcs, f'{arm.lower()}arm_{stage.lower()}_analog')
            if val != '':
                [f, tf_ana] = load_foton_export_tf(self.dpath(val))
                tf_ana_interp = np.interp(frequencies, f, tf_ana)

        # Get the CALCS actuation output matrix (different then the DARM output
        # matrix)
        # TODO: this is -1 for X and I think +1 for Y; why is this?
        if hasattr(self.calcs, f'{arm.lower()}arm_output_matrix'):
            idx = ['top', 'uim', 'pum', 'tst'].index(stage.lower())
            out_mtrx = getattr(self.calcs,
                               f'{arm.lower()}arm_output_matrix')[idx]
        else:
            out_mtrx = 0

        # put in the gain and the DARM feedback sign
        # FOTON export is in m/N, here we apply the N/ct from the model
        # assuming that this has been replicated perfectly in CALCS.
        if (getattr(self.actuation, f'{arm.lower()}arm') is not None and
                stage.lower() in ['uim', 'pum', 'tst']):
            if ((stage.lower() in ['uim', 'pum'] and
                    hasattr(getattr(self.actuation, f'{arm.lower()}arm'),
                            f'{stage.lower()}_npa') and
                    getattr(getattr(self.actuation, f'{arm.lower()}arm'),
                            f'{stage.lower()}_npa') != '') or
                    (stage.lower() == 'tst' and
                        hasattr(getattr(self.actuation, f'{arm.lower()}arm'),
                                f'{stage.lower()}_npv2') and
                        getattr(getattr(self.actuation, f'{arm.lower()}arm'),
                                f'{stage.lower()}_npv2') != '')):
                act_arm_model = getattr(self.actuation, f'{arm.lower()}arm')
                if stage.lower() == 'uim':
                    tf_ana_interp *= act_arm_model.uim_dc_gain_Npct()
                elif stage.lower() == 'pum':
                    tf_ana_interp *= act_arm_model.pum_dc_gain_Npct()
                else:
                    tf_ana_interp *= act_arm_model.tst_dc_gain_Npct()

        # apply CALCS digital gain, default is 1
        act_gain = getattr(self, f'{arm}arm_{stage.lower()}_analog_gain', 1.0)
        tf_ana_interp *= act_gain

        out_tf = (out_mtrx *
                  tf_dig_interp *
                  tf_ana_interp)

        return out_tf

    def drivealign_out_to_darm_displacement(self, frequencies, arm, stage):
        """
        CALCS representation of the SUS_EXC DRIVEALIGN point to DARM
        displacement transfer function.

        Units are m/ct = (N/ct) * (m/N)


        The gain (N/ct) of the CALCS model is assumed to be exactly the same
        as what is in the reference model, so that is used.

        If there is anything in the "coiloutf" key-value pair, then this
        transfer function is also used, but this would be used only in
        special circumstances.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies
        arm : `str`
            arm for the calculation, 'x' or 'y'
        stage : `str`
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'

        Returns
        -------
        tf : `complex128`, array-like
            CALCS approximation of the DRIVEALIGN output to DARM displacement
        """

        if (getattr(self.actuation, f'{arm.lower()}arm') is None):
            raise ValueError('Must provide a valide arm object')

        act_arm_model = getattr(self.actuation, f'{arm.lower()}arm')

        tf = (act_arm_model.darm_feedback_sign *
              self.drivealign_out_to_longitudinal_displacement(
                  frequencies, arm, stage))

        return tf

    def compute_actuation_single_stage(self, frequencies, arm, stage):
        """
        A_n,calcs

        CALCS approximation of a specific arm, stage from input to SUS ISCINF
        to DARM (and therefore does not include the OMC to CALCS delay)
        displacement. See T1900169


        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the CALCS approximated DARM actuation
        arm : `str`
            arm for the calculation, 'x' or 'y'
        stage : `str`
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'

        Returns
        -------
        calcs_stage_actuation : `complex128`, array-like
            CALCS approximation of SUS model ISCINF input to DARM displacement
            transfer function
        """

        # Assume the same as in-loop DARM output matrix
        # only used as determining if this is non-zero
        output_matrix = self.actuation.darm_output_matrix_values()

        # initialize zeros for the output
        calcs_stage_actuation = np.zeros(len(frequencies), dtype='complex128')

        arm_idx = ['x', 'y'].index(arm.lower())
        stage_idx = ['top', 'uim', 'pum', 'tst'].index(stage.lower())
        if (output_matrix[arm_idx, stage_idx] != 0.0):
            assert stage_idx > 0, 'top has not been implemented'

            # Model SUS filters
            arm_obj = getattr(self.actuation, f'{arm}arm')
            dig_filt = arm_obj.sus_digital_filters_response(frequencies)

            # This is the CALCS DARM displacement for this arm and stage
            drivealign_out_to_disp = self.drivealign_out_to_darm_displacement(
                frequencies, arm=arm, stage=stage)

            calcs_stage_actuation = (dig_filt[stage_idx - 1] *
                                     drivealign_out_to_disp)

        return calcs_stage_actuation

    def stage_super_actuator(self, frequencies, stage='TST'):
        """
        Compute the super actuator transfer function for a specific stage as
        estimated by the CALCS model.
        In this case, the "stage super actuator" is created by choosing a
        specific stage and then for each QUAD, the stage transfer function to
        DARM is summed together.

        This transfer function is from DARM_CTRL to meters sensed by the IFO.
        Note that the sign of the DARM_ERR signal is dependent upon which arm
        is under control.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        stage : `str`, optional
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the actuation function
        """

        # Assume the same as in-loop DARM output matrix
        output_matrix = self.actuation.darm_output_matrix_values()

        super_actuation = np.zeros(len(frequencies), dtype='complex128')

        stage_idx = ['top', 'uim', 'pum', 'tst'].index(stage.lower())
        for arm_idx, arm in enumerate(['x', 'y']):
            if output_matrix[arm_idx, stage_idx] != 0.0:
                assert stage_idx > 0, 'top has not been implemented'

                single_stage = self.compute_actuation_single_stage(
                        frequencies, arm=arm, stage=stage.upper())

                super_actuation += (
                    output_matrix[arm_idx, stage_idx] * single_stage)

        return super_actuation

    def arm_super_actuator(self, frequencies, arm='x'):
        """
        Compute the super actuator transfer function for a specific arm.
        In this case, the "arm super actuator" is created by choosing a
        specific arm and then for each QUAD, the arm transfer function to
        DARM is summed together.

        This transfer function is from DARM_CTRL to meters sensed by the IFO.
        Note that the sign of the DARM_ERR signal is dependent upon which arm
        is under control.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        arm : `str`, optional
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'

        Returns
        -------
        super_actuation : `complex128`, array-like
            transfer function response of the actuation function
        """

        # Assume the same as in-loop DARM output matrix
        output_matrix = self.actuation.darm_output_matrix_values()

        super_actuation = np.zeros(len(frequencies), dtype='complex128')

        arm_idx = ['x', 'y'].index(arm.lower())
        for stage_idx, stage in enumerate(['top', 'uim', 'pum', 'tst']):
            if output_matrix[arm_idx, stage_idx] != 0.0:
                assert stage_idx > 0, 'top has not been implemented'

                single_stage = self.compute_actuation_single_stage(
                        frequencies, arm=arm, stage=stage.upper())

                super_actuation += (
                    output_matrix[arm_idx, stage_idx] * single_stage)

        return super_actuation

    def gds_actuation_correction(self, frequencies, stage, daqdownsample=True):
        """
        Compute the correction to the CAL-CS output for GDS. Note that this
        implicitly assumes that the front end digital filters in CALCS is the
        same as that in the SUS path!

        This is also more complicated than sensing because there is a single
        channel for a given stage that accounts for BOTH arms; e.g.,
        ${IFO}:CAL-DELTAL_CTRL_${STAGE}_DBL_DQ takes input from both x and y
        arms using an output matrix.

        This is (A_{xn}+A_{yn})/(A_{xn,calcs}+A_{yn,calcs})/(F*delay)
        where n is the stage, calcs denotes FOTON implementation of an actuator
        F is the DAQ downsampling filter (on by default), and delay of the 1
        16384 clock cycle delay going from OMC to CALCS

        See T1900169, definition of delta A_n

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the GDS sensing correction
        stage : `str`
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'
        daqdownsample : `bool`, optional
            When True (default), the 16k-4k daqdownsample is applied in
            GDS actuation correction

        Returns
        -------
        correction : `complex128`, array-like
            transfer function response of the correction
        """

        # ideal model sum of actuators for A_{xn} and A_{yn}
        # this DOES NOT include the OMC to SUS delay
        A_n = self.actuation.stage_super_actuator(frequencies,
                                                  stage=stage.upper())

        # CALCS approximate for A_{xn,calcs} and A_{yn,calcs}
        # this DOES include the OMC to CALCS delay
        A_n_calcs = self.stage_super_actuator(frequencies,
                                              stage=stage.upper())

        # 1 clock cycle delay transfer function for OMC to CALCS
        omc_to_calcs_response = (
            signal.dfreqresp(digital_delay_filter(1, 16384),
                             2.0*np.pi*frequencies/16384)[1])

        if daqdownsample:
            # DAQ downsampling filters are applied so we need to account
            # for this
            daqdownsampling = signal.dfreqresp(
                daqdownsamplingfilters(16384, 4096, 'biquad', 'v3'),
                2.0*np.pi*frequencies/16384)[1]
        else:
            daqdownsampling = 1

        # This is \delta A_n
        # note that A_n already includes the OMC to SUS response
        correction = ((A_n / A_n_calcs) *
                      (1 / (daqdownsampling * omc_to_calcs_response)))

        return correction

    def calcs_darm_actuation(self, frequencies):
        """
        Compute the CALCS approximated DARM actuation. This method implicitly
        assumes that the CALCS DARM output matrix and actuation digital
        filtering matches the installed in-loop DARM ouput matrix and actuation
        digital filtering.

        See T1900169, definition of A_CALCS

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the CALCS approximated DARM actuation

        Returns
        -------
        calcs_actuation : `complex128`, array-like
            Residual phase after removing the simulated delay
        """

        # Start with zeros
        calcs_actuation = np.zeros(len(frequencies), dtype='complex128')

        # This has the OMC to CALCS time delay
        for stage_idx, stage in enumerate(['top', 'uim', 'pum', 'tst']):
            calcs_actuation += self.stage_super_actuator(
                frequencies, stage=stage.upper())

        return calcs_actuation

    def gds_sensing_correction(self, frequencies):
        """
        Compute 1 / C_corr = C_foton / C see T1900169.

        For the sensing path, GDS sensing correction = 1 / C_corr. This is
        to be multiplied into DELTAL_RESIDUAL

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the GDS sensing correction

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the correction
        """

        return 1 / self.C_corr(frequencies)

    def A_corr(self, frequencies, daqdownsample=True):
        """
        Compute total DARM actuator A_corr = A / A_foton (aka delta A),
        see T1900169.

        This is more complicated than computing C_corr, since the GDS actuation
        correction is applied to each stage separately, so we cannot just sum
        the corrections and multiply by a common factor. Instead, we compute
        the modeled (true) DARM actuation, and divide by the CALCS
        (approximate) DARM actuation. We also account for the OMC to CALCS
        delay and the possibility for DAQ downsampling.

        DAQ downsample is an option here because if you want to create
        delta L_ctrl from ${IFO}:CAL-DELTAL_CTRL_DBL_DQ, then we need to
        use daqdownsampling = True because the channel that is read has
        been downsampled. When computing the delay between CAL-DELTAL_CTRL
        and CAL-DELTAL_RESIDUAL to improve CAL-DELTAL_EXTERNAL, however, one
        does not need daqdownsampling (see discussion in sec III of
        T1900169).

        Both products are using the CALCS copy of DARM_CTRL and therefore need
        the delay (OMC to CALCS)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the GDS sensing correction
        daqdownsample : `bool`, optional
            When True (default), the 16k-4k daqdownsample is applied in A_corr

        Returns
        -------
        A_corr : `complex128`, array-like
            transfer function response of the correction
        """

        if daqdownsample:
            # DAQ downsampling filters are applied so we need to account
            # for this
            daqdownsampling = signal.dfreqresp(
                daqdownsamplingfilters(16384, 4096, 'biquad', 'v3'),
                2.0*np.pi*frequencies/16384)[1]
        else:
            daqdownsampling = 1

        # we will need to apply a delay because when it is inverted in GDS
        # this will become an advance (the correct thing to do)
        omc_to_calcs_response = (
            signal.dfreqresp(digital_delay_filter(1, 16384),
                             2.0*np.pi*frequencies/16384)[1])

        A = self.actuation.compute_actuation(frequencies)
        A_calcs = self.calcs_darm_actuation(frequencies)

        A_corr = ((A / A_calcs) *
                  (1 / (daqdownsampling * omc_to_calcs_response)))

        return A_corr

    def sensing_actuation_delay(self, frequencies, clock=False):
        """
        Compute the delay based on the GDS sensing and actuation correction in
        units of seconds delay or (if clock=True) in units of 16384
        samples/sec, i.e., 1 16384 Hz clock delay = 6.1035e-5 s. Positive
        indicates a delay, negative indicates an advance

        See T1900169, sec III "method 1"

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to model the phase delay
        clock : `boolean`, optional
            if clock is True, then the output delay_fit and residual will
            be computed and rounded to the nearest number of integer clock
            cycles (16384)

        Returns
        -------
        delay : float
            Units of seconds or 16384 Hz clock cycles (if clock=True)
        """

        # We use the C_corr = \delta C and A_corr = \delta A values from
        # T1900169, but note that we TURN OFF the DAQ downsample here
        # because the path that is summed in CALCS does not have the DAQ
        # downsampling
        C_corr = self.C_corr(frequencies)
        A_corr = self.A_corr(frequencies, daqdownsample=False)

        delay = -np.angle(C_corr*A_corr) / (2.0*np.pi*frequencies)

        if clock is True:
            delay = delay * 16384

        return delay

    def deltal_ext_whitening(self, frequencies):
        """
        Compute the interpolated transfer function of the DELTAL_EXTERNAL
        export from FOTON

        If no export file is provided in the configuration, this method
        will return an array of ones

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the interpolation

        Returns
        -------
        whitening_interp : `complex128`, array-like
            transfer function response of the calibration
        """

        if (hasattr(self.calcs, 'foton_deltal_whitening_tf') and
                getattr(self.calcs, 'foton_deltal_whitening_tf') != ''):
            foton_freq, foton_tf = load_foton_export_tf(
                self.dpath(self.calcs.foton_deltal_whitening_tf))
            whitening_interp = np.interp(frequencies,
                                         foton_freq,
                                         foton_tf)
        else:
            whitening_interp = np.ones(len(frequencies), dtype='complex128')

        return whitening_interp

    def calcs_dtt_calibration(self, frequencies, include_whitening=True,
                              strain_calib=False, save_to_file=None, fmt=None):
        """
        Compute the calibration transfer function, dL_pyDARM / dL_CALCS
        for the main control room calibrated sensitivity curve. One can save
        this data to a file with the needed frequency, dB magnitude, and
        degrees phase columns, and apply it to a DTT template.

        See T1900169 section IV

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the CAL_DELTAL_EXTERNAL DTT
            calibration
        include_whitening : `bool`, optional
            if the whitening filter is on (default), then we'd like to remove
            its effect so by default this divides out the whitening filter
        strain_calib : `bool`, optional
            the output defaults to dL_pyDARM/dL_CALCS, so that we can recalibrate
            CAL-DELTAL_EXTERNAL. If this option is True, then the output is
            h_pyDARM/dL_CALCS
        save_to_file : `str`, optional
            Filename (ASCII) to save the data from this result. Note:
            the default file columns are
            <frequency> <magnitude (dB)> <phase (deg)>
        fmt : `str`, optional
            if save_to_file is used, then this is the output format. Options
            are 'dB,deg' (default), 'mag,deg', 'dB,rad', 'mag,rad', or 're,im'.
            If none is given, then the default is used

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the calibration
        """

        # This is C_r*delay*(opt resp)/(opt resp)_foton
        C_corr = self.C_corr(frequencies)

        # Remember that A_corr should not have a DAQ downsample since
        # the path that is summed does not have the DAQ downsample in it
        A_corr = self.A_corr(frequencies, daqdownsample=False)

        # Whitening filter
        whitening_interp = self.deltal_ext_whitening(frequencies)

        if not include_whitening:
            whitening_interp = 1

        # Foton delay filter
        foton_freq, foton_tf = load_foton_export_tf(
            self.dpath(self.calcs.foton_delay_filter_tf))
        delay_interp = np.interp(frequencies, foton_freq, foton_tf)

        # Model loop gain CDA
        G = self.compute_darm_olg(frequencies)

        calib = ((1 / whitening_interp) *
                 (1 / C_corr) *
                 ((1.0 + G) /
                  (1.0 + G*delay_interp/(C_corr*A_corr))))

        if strain_calib:
            calib /= self.sensing.mean_arm_length()

        # Save to file and use the fmt options if given
        # default is to save as dB magnitude and degrees phase
        if save_to_file is not None:
            col1 = frequencies
            col2 = 20.0*np.log10(np.abs(calib))
            col3 = np.angle(calib, deg=True)

            if fmt is not None:
                fmt_opt = fmt.split(',')
                if len(fmt_opt) != 2:
                    raise ValueError('Wrong formatting for fmt')
                if fmt_opt[0] == 'mag':
                    col2 = np.abs(calib)
                elif fmt_opt[0] == 're':
                    col2 = np.real(calib)
                if fmt_opt[1] == 'rad':
                    col3 = np.angle(calib, deg=False)
                elif fmt_opt[1] == 'im':
                    col3 = np.imag(calib)

            np.savetxt(save_to_file,
                       np.array([col1,
                                 col2,
                                 col3]).T,
                       fmt='%.7g')

        return calib

    def deltal_ext_pcal_correction(self, frequencies, **kwargs):
        """
        Compute the calibration transfer function, dL_pyDARM / dL_Pcal
        for determining the response function systematic error

        See T1900169 section V

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the CAL_DELTAL_EXTERNAL DTT
            calibration
        kwargs : optional
            values are passed to `calcs.calcs_dtt_calibration()` and
            `pcal.compute_pcal_correction()`

        Returns
        -------
        tf : `complex128`, array-like
            transfer function resposne of the correction factor
        """

        endstation = kwargs.pop('endstation', False)
        include_dewhitening = kwargs.pop('include_dewhitening', True)

        deltal_ext = self.calcs_dtt_calibration(frequencies, **kwargs)
        pcal_corr = self.pcal.compute_pcal_correction(
            frequencies,
            endstation=endstation,
            include_dewhitening=include_dewhitening)

        tf = deltal_ext / pcal_corr

        return tf

    def compute_epics_records(self, endstation=False):
        """Generate EPICS values for CALCS front-end model

        This is a LIGO specific function for generating front end
        calibration EPICS records.

        See T1700106-v10 and G1601472 for additional details

        Parameters
        ----------
        endstation : `bool`, optional
            When false (default), the correction is computed for CAL-CS
            PCAL channel, which includes 1 16k clock cycle delay that we must
            compensate (undo). Otherwise, when true, the correction is computed
            at the end station, which does not include 1 16k clock cycle delay.

        Returns
        -------
        out : dict
            Dictionary of EPICS records

        """
        f_uim = self.calcs.cal_line_sus_uim_frequency
        f_pum = self.calcs.cal_line_sus_pum_frequency
        f_tst = self.calcs.cal_line_sus_tst_frequency
        f_pcal1 = self.calcs.cal_line_sus_pcal_frequency
        f_pcal2 = self.calcs.cal_line_sens_pcal_frequency
        f_pcal3 = self.calcs.cal_line_high_pcal_frequency
        f_pcalx_cmp = self.calcs.cal_line_cmp_pcalx_frequency
        f_pcaly_cmp = self.calcs.cal_line_cmp_pcaly_frequency

        freq = np.array(
            [f_pcal1, f_uim, f_pum, f_tst, f_pcal2, f_pcal3, f_pcalx_cmp,
             f_pcaly_cmp])

        R = self.compute_response_function(freq)

        daqdownsampling = daqdownsamplingfilters(2**14, 2**9, 'biquad', 'v3')
        daqdownsampling_response = signal.dfreqresp(daqdownsampling,
                                                    2.0*np.pi*freq/2**14)[1]

        # Pcal corrections
        # See T1700106
        pcal_correction = self.pcal.compute_pcal_correction(freq, endstation)
        PCAL_LINE1_CORRECTION = (
            pcal_correction[0] * np.exp(-2*np.pi*1j*freq[0]/16384))
        PCAL_LINE2_CORRECTION = (
            pcal_correction[4] * np.exp(-2*np.pi*1j*freq[4]/16384))
        PCAL_LINE3_CORRECTION = (
            pcal_correction[5] * np.exp(-2*np.pi*1j*freq[5]/16384))

        # We divide out the pcal reference arm sign since we know which arm
        # the next two values will be assigned to. This makes the next two
        # values constant no matter what arm the PCal calibration lines are
        # assigned to
        PCAL_X_COMPARE_CORRECTION = (
            pcal_correction[6] / self.pcal.ref_pcal_2_darm_act_sign *
            np.exp(-2*np.pi*1j*freq[6]/16384))
        PCAL_Y_COMPARE_CORRECTION = (
            -1 * pcal_correction[7] / self.pcal.ref_pcal_2_darm_act_sign *
            np.exp(-2*np.pi*1j*freq[7]/16384))

        # this does not include any OMC to SUS model jump delay
        uim_actuation_epics = self.actuation.stage_super_actuator_drivealign(
            freq, stage='UIM')
        pum_actuation_epics = self.actuation.stage_super_actuator_drivealign(
            freq, stage='PUM')
        tst_actuation_epics = self.actuation.stage_super_actuator_drivealign(
            freq, stage='TST')

        # super actuator stage
        # this does include OMC to SUS model jump delay
        uim = self.actuation.stage_super_actuator(freq, stage='UIM')
        pum = self.actuation.stage_super_actuator(freq, stage='PUM')
        tst = self.actuation.stage_super_actuator(freq, stage='TST')

        # Actuator EPICS
        # Need to divide out a DAQ downsampling filter from the reference A
        # values because channels sampled at 512 Hz have an AA filter applied
        # See T1700106
        SUS_LINE1_REF_INVA_UIM_RESPRATIO = (
            (1.0/(uim_actuation_epics[1]/daqdownsampling_response[1])) *
            (1.0/R[0]) * R[1] * np.exp(2*np.pi*1j*freq[1]/16384))
        SUS_LINE2_REF_INVA_PUM_RESPRATIO = (
            (1.0/(pum_actuation_epics[2]/daqdownsampling_response[2])) *
            (1.0/R[0]) * R[2] * np.exp(2*np.pi*1j*freq[2]/16384))
        SUS_LINE3_REF_INVA_TST_RESPRATIO = (
            (1.0/(tst_actuation_epics[3]/daqdownsampling_response[3])) *
            (1.0/R[0]) * R[3] * np.exp(2*np.pi*1j*freq[3]/16384))

        # Need to add a phase delay to the SUS excitation because the
        # SUS channel acquires a phase delay from the model jump
        # to CAL-CS and the calculation expects no delay (as in GDS).
        # Note that if a replica SUS excitation is used in the CAL-CS model
        # then these should actually be set to zero.
        SUS_LINE1_SUS_DEMOD_PHASE = (
            np.angle(np.exp(-2*np.pi*1j*freq[1]/16384), deg=True))
        SUS_LINE2_SUS_DEMOD_PHASE = (
            np.angle(np.exp(-2*np.pi*1j*freq[2]/16384), deg=True))
        SUS_LINE3_SUS_DEMOD_PHASE = (
            np.angle(np.exp(-2*np.pi*1j*freq[3]/16384), deg=True))

        # Compute the sensing function without the optical response but need to
        # include the optical gain
        coupled_cavity = self.sensing.optical_response(
            self.sensing.coupled_cavity_pole_frequency,
            self.sensing.detuned_spring_frequency,
            self.sensing.detuned_spring_q,
            pro_spring=self.sensing.is_pro_spring)
        sensing_no_cavity_response = (
            self.sensing.compute_sensing(freq) /
            signal.freqresp(coupled_cavity, 2.0*np.pi*freq)[1])
        PCAL_LINE2_REF_C_NOCAVPOLE = sensing_no_cavity_response[4]
        PCAL_LINE1_REF_C_NOCAVPOLE = sensing_no_cavity_response[0]

        # DARM digital filtering
        darm_digital_filter_response = self.digital.compute_response(freq)
        PCAL_LINE2_REF_D = darm_digital_filter_response[4]
        PCAL_LINE1_REF_D = darm_digital_filter_response[0]

        # Actuation
        PCAL_LINE2_REF_A_TST = tst[4]
        PCAL_LINE2_REF_A_PUM = pum[4]
        PCAL_LINE2_REF_A_UIM = uim[4]
        PCAL_LINE1_REF_A_TST = tst[0]
        PCAL_LINE1_REF_A_PUM = pum[0]
        PCAL_LINE1_REF_A_UIM = uim[0]

        # Output - see T1700106
        return {
            'CAL-CS_TDEP_PCAL_LINE1_CORRECTION_REAL': PCAL_LINE1_CORRECTION.real,
            'CAL-CS_TDEP_PCAL_LINE1_CORRECTION_IMAG': PCAL_LINE1_CORRECTION.imag,
            'CAL-CS_TDEP_SUS_LINE3_REF_INVA_TST_RESPRATIO_REAL': SUS_LINE3_REF_INVA_TST_RESPRATIO.real,  # noqa E501
            'CAL-CS_TDEP_SUS_LINE3_REF_INVA_TST_RESPRATIO_IMAG': SUS_LINE3_REF_INVA_TST_RESPRATIO.imag,  # noqa E501
            'CAL-CS_TDEP_SUS_LINE2_REF_INVA_PUM_RESPRATIO_REAL': SUS_LINE2_REF_INVA_PUM_RESPRATIO.real,  # noqa E501
            'CAL-CS_TDEP_SUS_LINE2_REF_INVA_PUM_RESPRATIO_IMAG': SUS_LINE2_REF_INVA_PUM_RESPRATIO.imag,  # noqa E501
            'CAL-CS_TDEP_SUS_LINE1_REF_INVA_UIM_RESPRATIO_REAL': SUS_LINE1_REF_INVA_UIM_RESPRATIO.real,  # noqa E501
            'CAL-CS_TDEP_SUS_LINE1_REF_INVA_UIM_RESPRATIO_IMAG': SUS_LINE1_REF_INVA_UIM_RESPRATIO.imag,  # noqa E501
            'CAL-CS_TDEP_PCAL_LINE2_REF_C_NOCAVPOLE_REAL': PCAL_LINE2_REF_C_NOCAVPOLE.real,
            'CAL-CS_TDEP_PCAL_LINE2_REF_C_NOCAVPOLE_IMAG': PCAL_LINE2_REF_C_NOCAVPOLE.imag,
            'CAL-CS_TDEP_PCAL_LINE2_REF_D_REAL': PCAL_LINE2_REF_D.real,
            'CAL-CS_TDEP_PCAL_LINE2_REF_D_IMAG': PCAL_LINE2_REF_D.imag,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_TST_REAL': PCAL_LINE2_REF_A_TST.real,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_TST_IMAG': PCAL_LINE2_REF_A_TST.imag,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_PUM_REAL': PCAL_LINE2_REF_A_PUM.real,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_PUM_IMAG': PCAL_LINE2_REF_A_PUM.imag,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_UIM_REAL': PCAL_LINE2_REF_A_UIM.real,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_UIM_IMAG': PCAL_LINE2_REF_A_UIM.imag,
            'CAL-CS_TDEP_PCAL_LINE2_CORRECTION_REAL': PCAL_LINE2_CORRECTION.real,
            'CAL-CS_TDEP_PCAL_LINE2_CORRECTION_IMAG': PCAL_LINE2_CORRECTION.imag,
            'CAL-CS_TDEP_PCAL_LINE1_REF_C_NOCAVPOLE_REAL': PCAL_LINE1_REF_C_NOCAVPOLE.real,
            'CAL-CS_TDEP_PCAL_LINE1_REF_C_NOCAVPOLE_IMAG': PCAL_LINE1_REF_C_NOCAVPOLE.imag,
            'CAL-CS_TDEP_PCAL_LINE1_REF_D_REAL': PCAL_LINE1_REF_D.real,
            'CAL-CS_TDEP_PCAL_LINE1_REF_D_IMAG': PCAL_LINE1_REF_D.imag,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_TST_REAL': PCAL_LINE1_REF_A_TST.real,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_TST_IMAG': PCAL_LINE1_REF_A_TST.imag,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_PUM_REAL': PCAL_LINE1_REF_A_PUM.real,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_PUM_IMAG': PCAL_LINE1_REF_A_PUM.imag,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_UIM_REAL': PCAL_LINE1_REF_A_UIM.real,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_UIM_IMAG': PCAL_LINE1_REF_A_UIM.imag,
            'CAL-CS_TDEP_PCAL_LINE3_CORRECTION_REAL': PCAL_LINE3_CORRECTION.real,
            'CAL-CS_TDEP_PCAL_LINE3_CORRECTION_IMAG': PCAL_LINE3_CORRECTION.imag,
            'CAL-CS_TDEP_PCAL_X_COMPARE_CORRECTION_REAL': PCAL_X_COMPARE_CORRECTION.real,
            'CAL-CS_TDEP_PCAL_X_COMPARE_CORRECTION_IMAG': PCAL_X_COMPARE_CORRECTION.imag,
            'CAL-CS_TDEP_PCAL_Y_COMPARE_CORRECTION_REAL': PCAL_Y_COMPARE_CORRECTION.real,
            'CAL-CS_TDEP_PCAL_Y_COMPARE_CORRECTION_IMAG': PCAL_Y_COMPARE_CORRECTION.imag,
            'CAL-CS_TDEP_SUS_LINE1_SUS_DEMOD_PHASE': SUS_LINE1_SUS_DEMOD_PHASE,
            'CAL-CS_TDEP_SUS_LINE2_SUS_DEMOD_PHASE': SUS_LINE2_SUS_DEMOD_PHASE,
            'CAL-CS_TDEP_SUS_LINE3_SUS_DEMOD_PHASE': SUS_LINE3_SUS_DEMOD_PHASE,
        }



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from .model import Model
from .sensing import SensingModel
from .actuation import DARMActuationModel
from .pcal import PcalModel
from .utils import compute_digital_filter_response
from .plot import critique


class DigitalModel(Model):
    """DARM digital filter model object

    """

    def __init__(self, config):
        super().__init__(config, measurement='digital')

    def compute_response(self, frequencies):
        """
        Compute DARM digital controller frequency response

        Uses filter ZPK transfer function response from Foton file.

        Parameters
        ----------
        frequencies

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the digital SUS filter

        """

        # Emit error if trying to use the old format
        if ((hasattr(self, 'digital_filter_modules_in_use') and
             not hasattr(self, 'digital_filter_modules'))):
            raise KeyError('Using old name format for filter modules. Please'
                           ' check your configuration string/file and use the'
                           ' updated format to specify filter module'
                           ' parameters.')

        response = np.ones(len(frequencies), dtype='complex128')
        for n in range(len(self.digital_filter_bank)):
            if n == 0:
                tf, pfilt = compute_digital_filter_response(
                    self.dpath(self.digital_filter_file),
                    self.digital_filter_bank[n],
                    self.digital_filter_modules[n],
                    self.digital_filter_gain[n], frequencies, pfilt=None)
            else:
                tf = compute_digital_filter_response(
                    self.dpath(self.digital_filter_file),
                    self.digital_filter_bank[n],
                    self.digital_filter_modules[n],
                    self.digital_filter_gain[n], frequencies, pfilt=pfilt)[0]
            response *= tf
        return response


class DARMModel(Model):
    """
    DARM model object

    This is a class to set up the model for the DARM loop from a
    configuration file with all the information about where the data is stored

    """

    def __init__(self, config, sensing=None, actuation=None, digital=None, pcal=None):
        super().__init__(config)
        self.sensing = sensing or SensingModel(config)
        self.actuation = actuation or DARMActuationModel(config)
        self.digital = digital or DigitalModel(config)
        self.pcal = pcal
        if not self.pcal and 'pcal' in self._config:
            self.pcal = PcalModel(config)

    def compute_darm_olg(self, frequencies):
        """
        Compute the entire DARM open loop transfer function (see G1501518)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the sensing function

        """
        C_response = self.sensing.compute_sensing(frequencies)
        A_response = self.actuation.compute_actuation(frequencies)
        D_response = self.digital.compute_response(frequencies)

        return C_response * A_response * D_response

    def compute_response_function(self, frequencies, sensing_syserr=None,
                                  actuation_syserr_dict=None):
        """
        Compute the entire DARM response transfer function (see G1501518)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        sensing_syserr : `complex`, array-like, optional
            multiplicative factor to include relative sensing systematic error
        actuation_syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}}

        Returns
        -------
        tf : `complex128`, array-like
            transfer function of the DARM closed-loop response

        """

        C_response = self.sensing.compute_sensing(frequencies)
        A_response = self.actuation.compute_actuation(frequencies,
                                                      actuation_syserr_dict)
        D_response = self.digital.compute_response(frequencies)

        if sensing_syserr is not None:
            C_response *= sensing_syserr

        return (1.0/C_response + D_response * A_response)

    def compute_etas(self, frequencies, sensing_syserr=None,
                     actuation_syserr_dict=None):
        """
        Compute multiplicative scaling factor to the response function.
        This returns "eta_R_C", applying sensing systematic error only;
        "eta_R_A", applying actuation systematic error only; and "eta_R",
        applying both sensing and systematic errors.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        sensing_syserr : `complex`, array-like, optional
            multiplicative factor to include relative sensing systematic error
        actuation_syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}, 'yarm': {'PUM': `complex`, array-like}}

        Returns
        -------
        eta_R_c : `complex128`, array-like
            multiplicative scaling factor for the response function, applying the
            sensing systematic errors only
        eta_R_a : `complex128`, array-like
            multiplicative scaling factor for the response function, applying the
            actuation systematic errors only
        eta_R : `complex128`, array-like
            multiplicative scaling factor for the response function, applying both
            sensing and actuation systematic errors

        """

        R = self.compute_response_function(frequencies)

        eta_R_c = self.compute_response_function(frequencies, sensing_syserr=sensing_syserr) / R
        eta_R_a = self.compute_response_function(frequencies,
                                                 actuation_syserr_dict=actuation_syserr_dict) / R
        eta_R = self.compute_response_function(frequencies, sensing_syserr=sensing_syserr,
                                               actuation_syserr_dict=actuation_syserr_dict) / R
        return eta_R_c, eta_R_a, eta_R

    def plot(self, plot_selection='all', freq_min=0.1, freq_max=5000,
             filename=None, ifo='', label=None, style=None, ugf_start=10,
             ugf_end=1000, show=None, **kwargs):
        """
        Make DARM critique plots

        This method produces critique models for 1 or 2 models.

        Parameters
        ----------
        plot_selection : `str`, optional
            Select plot type, one of: 'all' (default), 'optical', 'actuation',
            'clg', 'olg', or 'digital'
        freq_min : `float`, optional
            start frequency
        freq_max : `float`, optional
            end frequency
        filename : `str`, optional
            if given, ALL generated graphs will be saved in one pdf
        ifo : `str`, optional
            if given with a model to plot, it will appear in the
            graph titles
        label : `str` list, optional
            FIXME: what should this be?
        style : `str`, optional
            one of the styles matplotlib has or a user filename with style
        ugf_start : `float`, optional
            start frequency used for the search
        ugf_end : `float`, optional
            end frequency used for the search
        show : `bool`, optional
            if true the plot(s) will show
        **kwargs : optional
            Matplotlib values passed to plots

        """

        critique(self, freq_min=freq_min, freq_max=freq_max, filename=filename, label=label,
                 plot_selection=plot_selection, ifo=ifo, show=show,
                 ugf_start=ugf_start, ugf_end=ugf_end, **kwargs)



from ._const import (
    CAL_CONFIG_ROOT,
    CAL_DATA_ROOT,
)


def add_args(parser):
    pass


def main(args):
    """print environment and exit

    """
    print(f"CAL_CONFIG_ROOT: {CAL_CONFIG_ROOT}")
    print(f"CAL_DATA_ROOT  : {CAL_DATA_ROOT}")
